from typing import List

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class CFUncertaintyRemove(ABAction):
    """
    ABAction to remove the uncertainty from one or multiple Characterization Factors.
    """

    icon = qicons.clear
    text = "Remove uncertainty"

    @staticmethod
    @exception_dialogs
    def run(method_name: tuple, char_factors: List[tuple]):
        # create a list of CF's of which the uncertainty dict is removed
        cleaned_cfs = []
        for cf in char_factors:
            # if there's no uncertainty dict, we may continue
            if not isinstance(cf[1], dict):
                continue

            # else replace the uncertainty dict with the float found in the amount field of said dict
            cleaned_cfs.append((cf[0], cf[1]["amount"]))

        # if the list is empty we can return
        if not cleaned_cfs:
            return

        # else write the cf's to the method
        method = bd.Method(method_name)
        method_dict = {cf[0]: cf[1] for cf in method.load()}

        for cf in cleaned_cfs:
            method_dict[cf[0]] = cf[1]

        method.write(list(method_dict.items()))



from typing import Optional

from qtpy import QtWidgets, QtCore
from qtpy.QtCore import Qt, QDate
from qtpy.QtGui import QIntValidator, QDoubleValidator
from qtpy.QtWidgets import QLineEdit, QDateEdit


class JSONDelegate(QtWidgets.QStyledItemDelegate):
    """
    JSON Editor that supports multiple types (int, float, string, date (string))
    Assumes that the
    """

    def __init__(self, type_column: Optional[int] = None, parent=None):
        """
        :param type_column: column index of the type column
        """
        super().__init__(parent)
        self.type_column = type_column

    def createEditor(self, parent, option, index):
        type_index = index.model().index(index.row(), self.type_column)
        value_type = index.model().data(type_index, Qt.DisplayRole)
        if value_type == "int":
            editor = QLineEdit(parent)
            editor.setValidator(QIntValidator())
        elif value_type == "float":
            editor = QLineEdit(parent)
            editor.setValidator(QDoubleValidator())
        elif value_type == "string":
            editor = QLineEdit(parent)
        elif value_type == "date":
            editor = QDateEdit(parent)
        else:
            editor = super().createEditor(parent, option, index)
        editor.setProperty("value_type", value_type)
        return editor

    def setEditorData(self, editor, index):
        current_text = index.model().data(index, Qt.DisplayRole)
        value_type = editor.property("value_type")
        if value_type in ["int", "float", "string"]:
            editor.setText(current_text)
        elif value_type == "date":
            editor.setDate(QDate.fromString(current_text, "yyyy-MM-dd"))
            editor.setCalendarPopup(True)
        else:
            editor.setText(current_text)

    def setModelData(
        self,
        editor: QtWidgets.QLineEdit,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ):
        """Take the editor, read the given value and set it in the model"""
        value_type = editor.property("value_type")
        if value_type == "int":
            try:
                value = int(editor.text())
                model.setData(index, value, QtCore.Qt.EditRole)
            except ValueError:
                pass
        elif value_type == "float":
            try:
                value = float(editor.text())
                model.setData(index, value, QtCore.Qt.EditRole)
            except ValueError:
                pass
        elif value_type == "string":
            value = editor.text()
            model.setData(index, value, QtCore.Qt.EditRole)
        elif value_type == "date":
            editor: QDateEdit
            value = editor.date().toPython().isoformat()
            model.setData(index, value, QtCore.Qt.EditRole)
        else:
            value = editor.text()
            model.setData(index, value, QtCore.Qt.EditRole)



from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

from .project_switch import ProjectSwitch


class ProjectDuplicate(ABAction):
    """
    Duplicate the current project to a new name.

    This method prompts the user to input a new name for duplicating the current project.
    It performs validation to ensure the new name is not empty and does not already exist.
    If the provided name is valid, the current project is duplicated to the new name, and
    the application switches to the newly created project.

    Args:
        name (str, optional): The name of the current project to duplicate. Defaults to the
                              currently active project.

    Steps:
    - If no name is provided, use the current project name.
    - Prompt the user for a new project name.
    - Return if the user cancels or provides an empty name.
    - Check if the new name already exists and show an error message if it does.
    - If the provided name is not the current project, set it as the current project.
    - Duplicate the project to the new name without switching to it.
    - Switch to the newly created project using the `ProjectSwitch` action.
    """

    icon = qicons.copy
    text = "Duplicate this project"
    tool_tip = "Duplicate the project"

    @staticmethod
    @exception_dialogs
    def run(name: str = None):
        if name is None:
            name = bd.projects.current

        new_name, ok = QtWidgets.QInputDialog.getText(
            application.main_window,
            "Duplicate current project",
            f"Duplicate project ({name}) to new name:"
            + " " * 10,
        )

        if not ok or not new_name:
            return

        if new_name in bd.projects:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible.",
                "A project with this name already exists.",
            )
            return

        if name != bd.projects.current:
            bd.projects.set_current(name, update=False)
        bd.projects.copy_project(new_name, switch=False)  # don't switch because it will auto-update bw2 projects
        ProjectSwitch.run(new_name)  # switch using the action instead



from qtpy import QtWidgets, QtCore
from activity_browser import actions
from activity_browser.ui import widgets


class ImpactCategoryHeader(QtWidgets.QWidget):

    def __init__(self, parent: QtWidgets.QWidget):
        """
        Initializes the ActivityHeader widget.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
        """
        super().__init__(parent)
        self.impact_category = parent.impact_category

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def sync(self):
        """
        Synchronizes the widget with the current state of the activity.
        """
        self.clear_layout()
        self.layout().addLayout(self.build_grid())

    def clear_layout(self, layout: QtWidgets.QLayout = None):
        layout = layout or self.layout()

        if layout is None:
            return

        while layout.count():
            item = layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()
            elif item.layout() is not None:
                self.clear_layout(item.layout())

    def build_grid(self) -> QtWidgets.QGridLayout:
        grid = QtWidgets.QGridLayout(self)
        grid.setContentsMargins(0, 5, 0, 5)
        grid.setSpacing(10)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        setup = [
            ("Name:", QtWidgets.QLabel(str(self.impact_category.name)),),
            ("Unit:", QtWidgets.QLabel(str(self.impact_category.metadata.get("unit", "Undefined"))),),
        ]

        # Arrange widgets for display as a grid
        for i, (title, widget) in enumerate(setup):
            grid.addWidget(widgets.ABLabel.demiBold(title, self), i, 1)
            grid.addWidget(widget, i, 2, 1, 4)

        return grid


class ImpactCategoryName(QtWidgets.QLineEdit):
    """
    A widget that displays and edits the name of the activity.
    """

    def __init__(self, parent: ImpactCategoryHeader):
        """
        Initializes the ActivityName widget.

        Args:
            parent (ActivityHeader): The parent widget.
        """
        super().__init__(str(parent.impact_category.name), parent)
        self.editingFinished.connect(self.change_name)

    def change_name(self):
        """
        Changes the name of the activity if it has been modified.
        """
        if self.text() == self.parent().impact_category.name:
            return
        # actions.ActivityModify.run(self.parent().activity, "name", self.text())


from logging import getLogger

from qtpy import QtWidgets
from qtpy.QtCore import Signal, SignalInstance

import bw2data as bd

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui import threading, widgets
from activity_browser.bwutils.importers import ABExcelImporter

log = getLogger(__name__)


class DatabaseImporterExcel(ABAction):
    """ABAction to open the DatabaseImportWizard"""

    text = "Import database from brightway excel format"
    tool_tip = "Import database from brightway excel format"

    @classmethod
    @exception_dialogs
    def run(cls):
        # get the path from the user
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent=application.main_window,
            caption='Choose brightway excel database to import',
            filter='excel spreadsheet (*.xlsx);; All files (*.*)'
        )
        if not path:
            return

        import_setup = ImportSetup(title="Import from Excel", context={"path": path})
        import_setup.exec_()


class ImportSetup(widgets.ABWizard):

    class ExtractPage(widgets.ABThreadedWizardPage):
        title = "Extracting Database"
        subtitle = "Extracting database from excel file"

        class Thread(threading.ABThread):
            loaded: SignalInstance = Signal(object)

            def run_safely(self, path: str):
                importer = ABExcelImporter(path)
                self.loaded.emit(importer)

        def initializePage(self, context: dict):
            """Start the download thread"""
            self.thread.start(context["path"])
            self.thread.loaded.connect(lambda i: context.__setitem__("importer", i))

        def nextPage(self) -> type[QtWidgets.QWizardPage] | None:
            return ImportSetup.DatabaseName

    class DatabaseName(widgets.ABWizardPage):
        title = "Database Name"
        subtitle = "Enter the name of the database you wish to create"

        def __init__(self, parent=None):
            super().__init__(parent)
            self.db_name_edit = widgets.DatabaseNameEdit(
                label="Set database name:",
                database_preset="",
            )
            self.db_name_edit.textChanged.connect(self.completeChanged)

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.db_name_edit)
            self.setLayout(layout)

        def isComplete(self):
            return bool(self.db_name_edit.text())

        def initializePage(self, context: dict):
            self.db_name_edit.setText(context["importer"].db_name)

        def finalize(self, context: dict):
            context["database_name"] = self.db_name_edit.text()

        def nextPage(self):
            return ImportSetup.DatabaseLink

    class DatabaseLink(widgets.ABWizardPage):
        title = "Link Databases"
        subtitle = "Link the imported database to existing databases"

        def __init__(self, parent=None):
            super().__init__(parent)
            layout = QtWidgets.QGridLayout(self)
            self.setLayout(layout)
            self.link_dict_edit = {}

        def isComplete(self):
            return True

        def initializePage(self, context: dict):
            # fetch the unlinked databases from the importer
            importer = context["importer"]
            link_dbs = set([exc["database"] for exc in importer.unlinked])
            layout = self.layout()

            for i, db in enumerate(link_dbs):
                if db == importer.db_name:
                    continue

                layout.addWidget(QtWidgets.QLabel(db), i, 0)

                drop_down = QtWidgets.QComboBox(self)
                drop_down.addItems(sorted(bd.databases))

                if db in bd.databases:
                    drop_down.setCurrentText(db)

                layout.addWidget(drop_down, i, 1)

                self.link_dict_edit[db] = drop_down


        def finalize(self, context: dict):
            context["linking_dict"] = {k: v.currentText() for k, v in self.link_dict_edit.items()}

        def nextPage(self):
            return ImportSetup.InstallPage

    class InstallPage(widgets.ABThreadedWizardPage):
        title = "Importing Database"
        subtitle = "Importing database from .xlsx file"

        class Thread(threading.ABThread):
            """Thread to handle the install process"""

            def run_safely(self, importer: ABExcelImporter, database_name: str, linking_dict: dict):
                """Download the ecoinvent release"""
                importer.automated_import(database_name, linking_dict)

        def initializePage(self, context: dict):
            """Start the download thread"""
            self.thread.start(context["importer"], context["database_name"], context["linking_dict"])

    pages = [ExtractPage, DatabaseName, DatabaseLink, InstallPage]



import os
import datetime
from typing import Optional, Any
from logging import getLogger

import arrow
import numpy as np
import bw2data as bd
import pandas as pd

from qtpy import QtGui, QtWidgets, QtCore
from qtpy.QtCore import QPoint, QRect, QSize, Qt, QTimer, Signal, Slot, SignalInstance
from qtpy.QtWidgets import QSizePolicy, QTableView

from activity_browser.settings import ab_settings
from activity_browser.ui.icons import qicons
from activity_browser.ui import delegates

from .dialogs import FilterManagerDialog, SimpleFilterDialog


log = getLogger(__name__)


class CustomHeader(QtWidgets.QHeaderView):
    """Header which has a filter button on each cell that can trigger a signal.

    Largely based on https://stackoverflow.com/a/30938728
    """

    clicked: SignalInstance = Signal(int, str)

    _x_offset = 0
    _y_offset = (
        0  # This value is calculated later, based on the height of the paint rect
    )
    _width = 18
    _height = 18

    def __init__(self, orientation=Qt.Horizontal, parent=None):
        super(CustomHeader, self).__init__(orientation, parent)
        self.setSectionsClickable(True)

        self.column_indices = []
        self.has_active_filters = []  # list of column indices that have filters active
        self.event_pos = None

    def paintSection(self, painter, rect, logical_index):
        """Paint the button onto the column header."""
        painter.save()
        super(CustomHeader, self).paintSection(painter, rect, logical_index)
        painter.restore()

        self._y_offset = int(rect.height() - self._width)

        if logical_index in self.column_indices:
            option = QtWidgets.QStyleOptionButton()
            option.rect = QRect(
                rect.x() + self._x_offset,
                rect.y() + self._y_offset,
                self._width,
                self._height,
            )
            option.state = (
                QtWidgets.QStyle.State_Enabled | QtWidgets.QStyle.State_Active
            )

            # put the filter icon onto the label
            if logical_index in self.has_active_filters:
                option.icon = qicons.filter
            else:
                option.icon = qicons.filter_outline
            option.iconSize = QSize(16, 16)

            # set the settings to a PushButton
            self.style().drawControl(QtWidgets.QStyle.CE_PushButton, option, painter)

    def mousePressEvent(self, event):
        index = self.logicalIndexAt(event.pos())
        if index in self.column_indices:
            x = self.sectionPosition(index)
            if (
                x + self._x_offset < event.pos().x() < x + self._x_offset + self._width
                and self._y_offset < event.pos().y() < self._y_offset + self._height
            ):
                # the button is clicked

                # set the position of the lower left point of the filter button to spawn a menu
                pos = QPoint()
                pos.setX(x + self._x_offset + self._width)
                pos.setY(self._y_offset + self._height)
                self.event_pos = pos

                # emit the column index and the button (left/right) pressed
                self.clicked.emit(index, str(event.button()).split(".")[-1])
            else:
                # pass the event to the header (for sorting)
                super(CustomHeader, self).mousePressEvent(event)
        else:
            # pass the event to the header (for sorting)
            super(CustomHeader, self).mousePressEvent(event)
        self.viewport().update()


class PandasModel(QtCore.QAbstractTableModel):
    """Abstract pandas table model adapted from
    https://stackoverflow.com/a/42955764.
    """

    HEADERS = []
    updated: SignalInstance = Signal()

    def __init__(self, df: pd.DataFrame = None, parent=None):
        super().__init__(parent)
        self._dataframe: Optional[pd.DataFrame] = df
        self.filterable_columns = None
        self.different_column_types = {}
        # The list of columns which should be editable by the builtin checkbox editor
        # The value of the dict holds whether the value should also be displayed as text
        self._checkbox_editors: dict[int, tuple[bool, Any, Any]] = {}
        self._columns: list[str] = []

    @property
    def columns(self) -> list[str]:
        if self._dataframe is not None:
            return self._dataframe.columns
        return []

    def rowCount(self, parent=None, *args, **kwargs):
        return 0 if self._dataframe is None else self._dataframe.shape[0]

    def columnCount(self, parent=None, *args, **kwargs):
        return 0 if self._dataframe is None else self._dataframe.shape[1]

    def data(self, index, role=Qt.DisplayRole):
        """
        Return value for table index based on a certain DisplayRole enum.

        More on DisplayRole enums: https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum
        """
        if not index.isValid():
            return None
        # instantiate value only in case of DisplayRole or ToolTipRole
        value = None
        tt_date_flag = False  # flag to indicate if value is datetime object and role is ToolTipRole
        if role in [Qt.DisplayRole, Qt.ToolTipRole, "sorting", Qt.EditRole]:
            value = self._dataframe.iat[index.row(), index.column()]
            if isinstance(value, np.float64):
                value = float(value)
            elif isinstance(value, bool):
                value = str(value)
            elif isinstance(value, np.int64):
                value = value.item()
            elif isinstance(value, tuple):
                value = str(value)
            elif isinstance(value, datetime.datetime) and (
                Qt.DisplayRole or Qt.ToolTipRole
            ):
                tz = datetime.datetime.now(datetime.timezone.utc).astimezone()
                time_shift = -tz.utcoffset().total_seconds()
                if role == Qt.ToolTipRole:
                    value = (
                        arrow.get(value)
                        .shift(seconds=time_shift)
                        .format("YYYY-MM-DD HH:mm:ss")
                    )
                    tt_date_flag = True
                elif role == Qt.DisplayRole:
                    value = arrow.get(value).shift(seconds=time_shift).humanize()

        # Handle checkbox editors
        # Checkbox editors can return two values for one cell: the usual display value
        # and a checked / not checked enum. It is useful to return both, when the
        # underlying data is not bool, but text to visualize eventual errors.
        if index.column() in self._checkbox_editors:
            if role == Qt.ItemDataRole.CheckStateRole:
                value = self._dataframe.iat[index.row(), index.column()]
                if isinstance(value, str):
                    log.error(f"Expected bool, received str: {value}!!")
                true_value = self._checkbox_editors[index.column()][1]
                # Convert the data to an appropriate value for the checkbox
                return Qt.CheckState.Checked if value == true_value else Qt.CheckState.Unchecked
            display_value = self._checkbox_editors[index.column()][0]
            if role == Qt.ItemDataRole.DisplayRole and not display_value:
                return None

        # immediately return value in case of DisplayRole or sorting
        if role == Qt.DisplayRole or role == "sorting":
            return value

        # in case of ToolTipRole and date, always show the full date
        if tt_date_flag and role == Qt.ToolTipRole:
            return value

        # in case of ToolTipRole, check whether content fits the cell
        if role == Qt.ToolTipRole:
            parent = self.parent()
            fontMetrics = parent.fontMetrics()

            # get the width of both the cell, and the text
            column_width = parent.columnWidth(index.column())
            text_width = fontMetrics.horizontalAdvance(str(value))
            margin = 10

            # only show tooltip if the text is wider then the cell minus the margin
            if text_width > column_width - margin:
                return value

        return None

    def flags(self, index):
        return Qt.ItemIsSelectable | Qt.ItemIsEnabled

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._dataframe.columns[section]
        elif orientation == Qt.Vertical and role == Qt.DisplayRole:
            return self._dataframe.index[section]
        return None

    def row_data(self, index: int) -> list:
        """Return the row at index as a list."""
        return self._dataframe.iloc[index, :].tolist()

    def to_clipboard(self, rows, columns, include_header: bool = False):
        """Copy the given rows and columns of the dataframe to clipboard"""
        self._dataframe.iloc[rows, columns].to_clipboard(
            index=False, header=include_header
        )

    def to_csv(self, path: str) -> None:
        """Store the dataframe as csv in the given path."""
        self._dataframe.to_csv(path)

    def to_excel(self, path: str) -> None:
        """Store the underlying dataframe as excel in the given path"""
        self._dataframe.to_excel(excel_writer=path)

    def sync(self, *args, **kwargs) -> None:
        """(Re)build the dataframe according to the given arguments."""
        self._dataframe = pd.DataFrame([], columns=self.HEADERS)

    @staticmethod
    def proxy_to_source(proxy: QtCore.QModelIndex) -> QtCore.QModelIndex:
        """Step from the QSortFilterProxyModel to the underlying PandasModel."""
        model = proxy.model()
        if not hasattr(model, "mapToSource"):
            return proxy  # Proxy is actually the PandasModel
        return model.mapToSource(proxy)

    @staticmethod
    def test_query_on_column(test_type: str, col_data: pd.Series, query) -> bool:
        """Compare query and col_data on test_type, return array with boolean test results."""
        if test_type == "equals":
            return col_data == query
        elif test_type == "does not equal":
            return col_data != query
        elif test_type == "contains":
            return col_data.str.contains(query, regex=False)
        elif test_type == "does not contain":
            return ~col_data.str.contains(query, regex=False)
        elif test_type == "starts with":
            return col_data.str.startswith(query)
        elif test_type == "does not start with":
            return ~col_data.str.startswith(query)
        elif test_type == "ends with":
            return col_data.str.endswith(query)
        elif test_type == "does not end with":
            return ~col_data.str.endswith(query)
        elif test_type == "=":
            return col_data.astype(float) == float(query)
        elif test_type == "!=":
            return col_data.astype(float) != float(query)
        elif test_type == ">=":
            return col_data.astype(float) >= float(query)
        elif test_type == "<=":
            return col_data.astype(float) <= float(query)
        elif test_type == "<= x <=":
            return (float(query[0]) <= col_data.astype(float)) & (
                col_data.astype(float) <= float(query[1])
            )
        else:
            log.warning("unknown filter type >{}<, assuming 'EQUALS'".format(test_type))
            return col_data == query

    def get_filter_mask(self, filters: dict) -> pd.Series:
        """Generate a filter mask of the dataframe based on the filters.

        Returns a pd.Series of boolean results (the mask).
        """
        # get the column name from index
        fc_rev = {v: k for k, v in self.filterable_columns.items()}

        all_mode = filters["mode"]
        all_mask = None
        # iterate over columns
        for col_idx, col_filters in filters.items():
            if col_idx == "mode":
                continue
            col_name = fc_rev[col_idx]
            col_data = self._dataframe[col_name]
            col_mode = col_filters.get("mode", False)
            col_mask = None
            # iterate over filters within column
            for col_filt in col_filters["filters"]:
                if self.different_column_types.get(col_name, False):
                    # this is a 'num' column
                    filt_type, query = col_filt
                    col_data_ = col_data
                else:
                    # this is a 'str' column
                    filt_type, query, case_sensitive = col_filt
                    if case_sensitive:
                        col_data_ = col_data.astype(str)
                    else:
                        col_data_ = col_data.astype(str).str.upper()
                        query = query.upper()

                # run the test
                new_mask = self.test_query_on_column(filt_type, col_data_, query)
                if not any(new_mask):
                    # no matches for this mask, let user know:
                    log.info(
                        "There were no matches for filter: {}: '{}'".format(
                            col_filt[0], col_filt[1]
                        )
                    )

                # create or combine new mask within column
                if isinstance(col_mask, pd.Series) and col_mode == "AND":
                    col_mask = col_mask & new_mask
                elif isinstance(col_mask, pd.Series) and col_mode == "OR":
                    col_mask = col_mask + new_mask
                else:
                    col_mask = new_mask

            # create or combine new mask on columns
            if isinstance(all_mask, pd.Series) and all_mode == "AND":
                all_mask = all_mask & col_mask
            elif isinstance(all_mask, pd.Series) and all_mode == "OR":
                all_mask = all_mask + col_mask
            else:
                all_mask = col_mask
        return all_mask

    def set_read_only(self, read_only: bool):
        """Interface function, to support editable models"""
        pass

    def is_read_only(self) -> bool:
        """Interface function, to support editable models"""
        return True

    def set_builtin_checkbox_delegate(self, column: int, show_text_value: bool,
                                      true_value: Any = True, false_value: Any = False):
        """
        Enables the builtin checkbox delegate for columns.
        Can be used on bool values only.
        As the underlying data can be bool or string, we provide the values to be
        stored as parameters.
        """
        self._checkbox_editors[column] = (show_text_value, true_value, false_value)


class ABSortProxyModel(QtCore.QSortFilterProxyModel):
    """Reimplementation to allow for sorting on the actual data in cells instead of the visible data.

    See this for context: https://github.com/LCA-ActivityBrowser/activity-browser/pull/1151
    """

    def lessThan(self, left: QtCore.QModelIndex, right: QtCore.QModelIndex) -> bool:
        """Override to sort actual data, expects `left` and `right` are comparable.

        If `left` and `right` are not the same type, we check if numerical and empty string are compared, if that is the
        case, we assume empty string == 0.
        Added this case for: https://github.com/LCA-ActivityBrowser/activity-browser/issues/1215
        """
        left_data = self.sourceModel().data(left, "sorting")
        right_data = self.sourceModel().data(right, "sorting")

        if not left_data and not right_data:
            return True
        if type(left_data) is type(right_data):
            return left_data < right_data

        # comparing Falsys with types
        if (isinstance(left_data, (int, float))
                and not right_data
        ):  # comparing left number with nothing, compare against '0' instead
            return left_data < 0
        if (isinstance(left_data, str)
                and not right_data
        ):  # comparing left str with nothing, compare against "" instead
            return left_data < ""  # note we use '>' instead of '<', content should be above empty fields
        if (isinstance(right_data, (int, float))
                and not left_data
        ):  # comparing right number with nothing, compare against '0' instead
            return 0 < right_data
        if (isinstance(right_data, str)
                and not left_data
        ):  # comparing right str with nothing, compare against "" instead
            return right_data < ""  # note we use '>' instead of '<', content should be above empty fields

        raise ValueError(
            f"Cannot compare {left_data} and {right_data}, incompatible types."
        )


class ABMultiColumnSortProxyModel(ABSortProxyModel):
    """Subclass of QSortFilterProxyModel to enable sorting on multiple columns.

    The main purpose of this subclass is to override def filterAcceptsRow().

    Subclass based on various ideas from:
    https://stackoverflow.com/questions/47201539/how-to-filter-multiple-column-in-qtableview
    http://www.dayofthenewdan.com/2013/02/09/Qt_QSortFilterProxyModel.html
    https://gist.github.com/dbridges/4732790
    """

    def __init__(self, parent=None):
        super(ABMultiColumnSortProxyModel, self).__init__(parent)

        # the filter mask, an iterable array with boolean values on whether or not to keep the row
        self.mask = None

        # metric to keep track of successful matches on filter
        self.matches = 0

        # custom filter activation
        self.activate_filter = False

    def set_filters(self, mask) -> None:
        self.mask = mask
        self.matches = 0
        self.activate_filter = True
        self.invalidateFilter()
        self.activate_filter = False
        log.info("{} filter matches found".format(self.matches))

    def clear_filters(self) -> None:
        self.mask = None
        self.invalidateFilter()

    def filterAcceptsRow(self, row: int, parent) -> bool:
        # check if self.activate_filter is enabled, else return True
        if not self.activate_filter:
            return True
        # get the right index from the mask
        matched = self.mask.iloc[row]
        if matched:
            self.matches += 1
        return matched


class ABDataFrameView(QtWidgets.QTableView):
    """Base class for showing pandas dataframe objects as tables."""

    ALL_FILTER = "All Files (*.*)"
    CSV_FILTER = "CSV (*.csv);; All Files (*.*)"
    TSV_FILTER = "TSV (*.tsv);; All Files (*.*)"
    EXCEL_FILTER = "Excel (*.xlsx);; All Files (*.*)"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setVerticalScrollMode(QTableView.ScrollPerPixel)
        self.setHorizontalScrollMode(QTableView.ScrollPerPixel)

        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)

        self.setWordWrap(True)
        self.setAlternatingRowColors(True)
        self.setSortingEnabled(True)

        self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setHighlightSections(False)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)

        self.verticalHeader().setDefaultSectionSize(22)  # row height
        self.verticalHeader().setVisible(False)
        # Use a custom ViewOnly delegate by default.
        # Can be overridden table-wide or per column in child classes.
        self.setItemDelegate(delegates.ViewOnlyDelegate(self))

        self.table_name = "LCA results"
        # Initialize attributes which are set during the `sync` step.
        # Creating (and typing) them here allows PyCharm to see them as
        # valid attributes.
        self.model: Optional[PandasModel] = None
        self.proxy_model: Optional[ABSortProxyModel] = None

    def rowCount(self) -> int:
        return 0 if self.model is None else self.model.rowCount()

    @Slot(name="updateProxyModel")
    def update_proxy_model(self) -> None:
        self.proxy_model = ABSortProxyModel(self)
        self.proxy_model.setSourceModel(self.model)
        self.proxy_model.setSortCaseSensitivity(Qt.CaseInsensitive)
        self.setModel(self.proxy_model)

    @Slot(name="exportToClipboard")
    def to_clipboard(self):
        """Copy dataframe to clipboard"""
        rows = list(range(self.model.rowCount()))
        cols = list(range(self.model.columnCount()))
        self.model.to_clipboard(rows, cols, include_header=True)

    def savefilepath(
        self, default_file_name: str, caption: str = None, file_filter: str = None
    ):
        """Construct and return default path where data is stored

        Uses the application directory for AB
        """
        safe_name = bd.utils.safe_filename(default_file_name, add_hash=False)
        caption = caption or "Choose location to save lca results"
        filepath, _ = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            caption=caption,
            dir=str(os.path.join(ab_settings.data_dir, safe_name)),
            filter=file_filter or self.ALL_FILTER,
        )
        # getSaveFileName can now weirdly return Path objects.
        return str(filepath) if filepath else filepath

    @Slot(name="exportToCsv")
    def to_csv(self):
        """Save the dataframe data to a CSV file."""
        filepath = self.savefilepath(self.table_name, file_filter=self.CSV_FILTER)
        if filepath:
            if not filepath.endswith(".csv"):
                filepath += ".csv"
            self.model.to_csv(filepath)

    @Slot(name="exportToExcel")
    def to_excel(self, caption: str = None):
        """Save the dataframe data to an excel file."""
        filepath = self.savefilepath(
            self.table_name, caption, file_filter=self.EXCEL_FILTER
        )
        if filepath:
            if not filepath.endswith(".xlsx"):
                filepath += ".xlsx"
            self.model.to_excel(filepath)

    @Slot(QtGui.QKeyEvent, name="copyEvent")
    def keyPressEvent(self, e):
        """Allow user to copy selected data from the table

        NOTE: by default, the table headers (column names) are also copied.
        """
        if e.modifiers() & Qt.ControlModifier:
            # Should we include headers?
            headers = e.modifiers() & Qt.ShiftModifier
            if e.key() == Qt.Key_C:  # copy
                selection = [
                    self.model.proxy_to_source(p) for p in self.selectedIndexes()
                ]
                rows = [index.row() for index in selection]
                columns = [index.column() for index in selection]
                rows = sorted(set(rows), key=rows.index)
                columns = sorted(set(columns), key=columns.index)
                self.model.to_clipboard(rows, columns, headers)


class ABFilterableDataFrameView(ABDataFrameView):
    """Filterable base class for showing pandas dataframe objects as tables.

    To use this table, the following MUST be set in the table model:
    - self.filterable_columns: dict
        --> these columns are available for filtering
        --> key is column name, value is column index

    To use this table, the following MUST be set in the table view:
    - self.header.column_indices = list(self.model.filterable_columns.values())
        --> If not set, no filter buttons will appear.
        --> Probably wise to set in a `if isinstance(self.model.filterable_columns, dict):`
        --> This variable must be set any time the columns of the table change

    To use this table, the following can be set in the table model:
    - self.different_column_types: dict
        --> these columns require a different filter type than 'str'
        --> e.g. self.different_column_types = {'col_name': 'num'}
    """

    FILTER_TYPES = {
        "str": [
            "contains",
            "does not contain",
            "equals",
            "does not equal",
            "starts with",
            "does not start with",
            "ends with",
            "does not end with",
        ],
        "str_tt": [
            "values in the column contain",
            "values in the column do not contain",
            "values in the column equal",
            "values in the column do not equal",
            "values in the column start with",
            "values in the column do not start with",
            "values in the column end with",
            "values in the column do not end with",
        ],
        "num": ["=", "!=", ">=", "<=", "<= x <="],
        "num_tt": [
            "values in the column equal",
            "values in the column do not equal",
            "values in the column are greater than or equal to",
            "values in the column are smaller than or equal to",
            "values in the column are between",
        ],
    }

    def __init__(self, parent=None):
        super().__init__(parent)

        self.header = CustomHeader()
        self.setHorizontalHeader(self.header)

        self.filters = None
        self.different_column_types = {}
        self.header.clicked.connect(self.header_filter_button_clicked)
        self.selected_column = 0

        # quick-filter setup:
        self.prev_quick_filter = {}
        self.debounce_quick_filter = QTimer()
        self.debounce_quick_filter.setInterval(300)
        self.debounce_quick_filter.setSingleShot(True)
        self.debounce_quick_filter.timeout.connect(self.quick_filter)

    def header_filter_button_clicked(self, column: int, button: str) -> None:
        self.selected_column = column
        # this function is separate from the context menu in case we want to add right-click options later
        if button == "LeftButton":
            self.header_context_menu()

    def header_context_menu(self) -> None:
        menu = QtWidgets.QMenu(self)
        menu.setToolTipsVisible(True)

        col_type = self.model.different_column_types.get(
            {v: k for k, v in self.model.filterable_columns.items()}[
                self.selected_column
            ],
            "str",
        )

        # quick-filter bar
        self.input_line = QtWidgets.QLineEdit()
        self.input_line.setFocusPolicy(Qt.StrongFocus)
        if col_type == "num":
            self.input_line.setValidator(QtGui.QDoubleValidator())
        search = QtWidgets.QToolButton()
        search.setIcon(qicons.search)
        search.clicked.connect(menu.close)
        quick_filter_layout = QtWidgets.QHBoxLayout()
        quick_filter_layout.addWidget(self.input_line)
        quick_filter_layout.addWidget(search)
        quick_filter_widget = QtWidgets.QWidget()
        quick_filter_widget.setLayout(quick_filter_layout)
        quick_filter_widget.setToolTip(
            "Filter this column on the input,\n"
            "press 'enter' or the search button to filter"
        )
        # write previous filter to the quick-filter input if we have one
        if prev_filter := self.prev_quick_filter.get(self.selected_column, False):
            self.input_line.setText(prev_filter[1])
        else:
            self.input_line.setPlaceholderText("Quick filter ...")
        self.input_line.textChanged.connect(self.debounce_quick_filter.start)
        self.input_line.returnPressed.connect(menu.close)
        QAline = QtWidgets.QWidgetAction(self)
        QAline.setDefaultWidget(quick_filter_widget)
        menu.addAction(QAline)

        # More filters submenu
        mf_menu = QtWidgets.QMenu(menu)
        mf_menu.setToolTipsVisible(True)
        mf_menu.setIcon(qicons.filter)
        mf_menu.setTitle("More filters")
        filter_actions = []
        for i, f in enumerate(self.FILTER_TYPES[col_type]):
            fa = QtWidgets.QAction(text=f)
            fa.setToolTip(self.FILTER_TYPES[col_type + "_tt"][i])
            fa.triggered.connect(self.simple_filter_dialog)
            filter_actions.append(fa)
        for fa in filter_actions:
            mf_menu.addAction(fa)
        menu.addMenu(mf_menu)
        # edit filters main menu
        filter_man = QtWidgets.QAction(qicons.edit, "Manage filters")
        filter_man.triggered.connect(self.filter_manager_dialog)
        filter_man.setToolTip("Open the filter management menu")
        menu.addAction(filter_man)
        # delete column filters option
        col_del = QtWidgets.QAction(qicons.delete, "Remove column filters")
        col_del.triggered.connect(self.reset_column_filters)
        col_del.setToolTip("Remove all filters on this column")
        menu.addAction(col_del)
        col_del.setEnabled(False)
        if isinstance(self.filters, dict) and self.filters.get(
            self.selected_column, False
        ):
            col_del.setEnabled(True)
        # delete all filters option
        all_del = QtWidgets.QAction(qicons.delete, "Remove all filters")
        all_del.triggered.connect(self.reset_filters)
        all_del.setToolTip("Remove all filters in this table")
        menu.addAction(all_del)
        all_del.setEnabled(False)
        if isinstance(self.filters, dict):
            all_del.setEnabled(True)

        # Show existing filters for column
        if isinstance(self.filters, dict) and self.filters.get(
            self.selected_column, False
        ):
            menu.addSeparator()
            active_filters_label = QtWidgets.QAction(
                qicons.filter, "Active column filters:"
            )
            active_filters_label.setEnabled(False)
            menu.addAction(active_filters_label)
            active_filters = []
            for filter_data in self.filters[self.selected_column]["filters"]:
                if filter_data[0] == "<= x <=":
                    q = " and ".join(filter_data[1])
                else:
                    q = filter_data[1]
                filter_str = ": ".join([filter_data[0], q])
                f = QtWidgets.QAction(text=filter_str)
                f.setEnabled(False)
                active_filters.append(f)
            for f in active_filters:
                menu.addAction(f)

        self.input_line.setFocus()
        loc = self.header.event_pos
        menu.exec_(self.mapToGlobal(loc))

    @Slot(name="updateProxyModel")
    def update_proxy_model(self) -> None:
        self.proxy_model = ABMultiColumnSortProxyModel(self)
        self.proxy_model.setSourceModel(self.model)
        self.proxy_model.setSortCaseSensitivity(Qt.CaseInsensitive)
        self.setModel(self.proxy_model)

    def quick_filter(self) -> None:
        # remove weird whitespace from input
        query = (
            self.input_line.text().translate(str.maketrans("", "", "\n\t\r")).strip()
        )

        # convert to filter
        col_name = {v: k for k, v in self.model.filterable_columns.items()}[
            self.selected_column
        ]
        if self.model.different_column_types.get(col_name):
            # column is type 'num'
            filt = ("=", query)
        else:
            # column is type 'str'
            filt = ("contains", query, False)
        # check if quick filter exists for this col, if so; remove from self.filters
        if prev_filter := self.prev_quick_filter.get(self.selected_column, False):
            self.filters[self.selected_column]["filters"].remove(prev_filter)

        # place the filter in self.prev_quick_filter for next quick filter on this column
        self.prev_quick_filter[self.selected_column] = filt

        # apply the right filters
        if query != "":
            # the query is not empty, add it to the filters and apply them
            self.add_filter(filt)
            self.apply_filters()
        elif len(self.filters[self.selected_column]["filters"]) > 0:
            # the query is empty, but there are still filters for this column, so apply the filters
            self.apply_filters()
        else:
            # the query is empty, and there are no more filters for this column, reset this filter.
            self.reset_column_filters()

    def filter_manager_dialog(self) -> None:
        # get right data
        column_names = self.model.filterable_columns

        # show dialog
        dialog = FilterManagerDialog(
            column_names=column_names,
            filters=self.filters,
            filter_types=self.FILTER_TYPES,
            selected_column=self.selected_column,
            column_types=self.model.different_column_types,
        )
        if dialog.exec_() == FilterManagerDialog.Accepted:
            # set the filters
            filters = dialog.get_filters
            if filters != self.filters:
                # the filters returned from the dialog are different, actually apply the filters
                rm = []
                for col, qf in self.prev_quick_filter.items():
                    # check if quickfilters exist for these columns, otherwise remove them
                    if (
                        filters.get(col, False) and qf not in filters[col]["filters"]
                    ) or not filters.get(col, False):
                        rm.append(col)
                for col in rm:
                    self.prev_quick_filter.pop(col)
                self.write_filters(filters)
                self.apply_filters()

    def simple_filter_dialog(self, preset_type: str = None) -> None:
        if not preset_type:
            preset_type = self.sender().text()

        # get right data
        column_name = {v: k for k, v in self.model.filterable_columns.items()}[
            self.selected_column
        ]
        col_type = self.model.different_column_types.get(column_name, "str")

        # show dialog
        dialog = SimpleFilterDialog(
            column_name=column_name,
            filter_types=self.FILTER_TYPES,
            column_type=col_type,
            preset_type=preset_type,
        )
        if dialog.exec_() == SimpleFilterDialog.Accepted:
            new_filter = dialog.get_filter
            # add the filter to existing filters
            if new_filter:
                self.add_filter(new_filter)
                self.apply_filters()

    def add_filter(self, new_filter: tuple) -> None:
        """Add a single filter to self.filters."""
        if isinstance(self.filters, dict):
            # filters exist
            all_filters = self.filters
            if all_filters.get(self.selected_column, False):
                # filters exist for this column
                all_filters[self.selected_column]["filters"].append(new_filter)
                if (
                    not all_filters[self.selected_column].get("mode", False)
                    and len(all_filters[self.selected_column]["filters"]) > 1
                ):
                    # a mode does not exist, but there are multiple filters
                    all_filters[self.selected_column]["mode"] = "OR"
            else:
                # filters don't yet exist for this column:
                all_filters[self.selected_column] = {"filters": [new_filter]}
        else:
            # no filters exist
            all_filters = {
                self.selected_column: {"filters": [new_filter]},
                "mode": "AND",
            }

        self.write_filters(all_filters)

    def write_filters(self, filters: dict) -> None:
        self.filters = filters

    def apply_filters(self) -> None:
        if self.filters:
            QtWidgets.QApplication.setOverrideCursor(Qt.WaitCursor)
            # only allow filters that are for columns that may be filtered on
            filters = {
                k: v
                for k, v in self.filters.items()
                if k in list(self.model.filterable_columns.values()) + ["mode"]
            }
            self.proxy_model.set_filters(self.model.get_filter_mask(filters))
            self.header.has_active_filters = list(filters.keys())
            QtWidgets.QApplication.restoreOverrideCursor()
        else:
            self.reset_filters()

    def reset_column_filters(self) -> None:
        """Reset all filters for this column."""
        QtWidgets.QApplication.setOverrideCursor(Qt.WaitCursor)
        f = self.filters
        if f.get(self.selected_column, False):
            f.pop(self.selected_column)
        if self.prev_quick_filter.get(self.selected_column, False):
            self.prev_quick_filter.pop(self.selected_column)
        self.write_filters(f)
        if len(self.filters) == 1 and self.filters.get("mode"):
            # the only thing in filters remaining is the mode --> there are no filters
            self.reset_filters()
        else:
            self.header.has_active_filters = list(self.filters.keys())
            self.apply_filters()
        QtWidgets.QApplication.restoreOverrideCursor()

    def reset_filters(self) -> None:
        """Reset all filters for this entire table."""
        QtWidgets.QApplication.setOverrideCursor(Qt.WaitCursor)
        self.write_filters(None)
        self.header.has_active_filters = []
        self.prev_quick_filter = {}
        self.proxy_model.clear_filters()
        QtWidgets.QApplication.restoreOverrideCursor()


class LCAResultsModel(PandasModel):
    def sync(self, df):
        self._dataframe = df.replace(np.nan, "", regex=True)
        self.updated.emit()


class LCAResultsTable(ABDataFrameView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.model = LCAResultsModel(parent=self)
        self.model.updated.connect(self.update_proxy_model)


class InventoryModel(PandasModel):
    def sync(self, df):
        self._dataframe = df
        # set the visible columns
        self.filterable_columns = {
            col: i for i, col in enumerate(self._dataframe.columns.to_list())
        }
        # set the columns te be defined as num (all except the first five for both biopshere and technosphere
        self.different_column_types = {
            col: "num"
            for i, col in enumerate(self._dataframe.columns.to_list())
            if i >= 5
        }
        self.updated.emit()


class InventoryTable(ABFilterableDataFrameView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.horizontalHeader().setStretchLastSection(True)

        self.model = InventoryModel(parent=self)
        self.model.updated.connect(self.update_proxy_model)
        self.model.updated.connect(self.update_filter_data)
        # below variables are required for switching between technosphere and biosphere tables
        self.showing = None
        self.filters_tec = None
        self.filters_bio = None

    def update_filter_data(self) -> None:
        if self.showing == "technosphere":
            self.filters = self.filters_tec
        else:
            self.filters = self.filters_bio

        # update the column header indices
        if isinstance(self.model.filterable_columns, dict):
            self.header.column_indices = list(self.model.filterable_columns.values())
        # apply the existing filters
        self.apply_filters()

    def write_filters(self, filters: dict) -> None:
        if self.showing == "technosphere":
            self.filters_tec = filters
        else:
            self.filters_bio = filters
        self.filters = filters


class ContributionModel(PandasModel):
    def sync(self, df, unit="relative share"):

        if "unit" in df.columns:
            # overwrite the unit col with 'relative share' if looking at relative results (except 3 'total' and 'rest' rows)
            df["unit"] = [""] * 3 + [unit] * (len(df) - 3)

        # drop any rows where all numbers are 0
        self._dataframe = df.loc[~(df.select_dtypes(include=np.number) == 0).all(axis=1)]
        self.updated.emit()


class ContributionTable(ABDataFrameView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.model = ContributionModel(parent=self)
        self.model.updated.connect(self.update_proxy_model)



import pickle
from logging import getLogger

from qtpy import QtCore, QtWidgets, QtGui

import bw2data as bd

from activity_browser import signals, application
from activity_browser.ui import icons

from activity_browser.ui.menu_bar import MenuBar

log = getLogger(__name__)


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        self.setWindowTitle("Activity Browser")
        self.setDockNestingEnabled(True)

        # Layout: extra items outside main layout
        self.menu_bar = MenuBar(self)
        self.setMenuBar(self.menu_bar)

        self.connect_signals()

    def sync(self):
        """
        Synchronizes the main window layout with the current Brightway2 project.

        This method clears existing panes, initializes default panes, and arranges them
        in the main window. Hidden panes are set to be invisible, and the first pane is
        raised to the top. The window title is updated to reflect the current project.

        Steps:
        - Clear all existing panes.
        - Create and add default panes as dock widgets.
        - Hide panes that are marked as hidden.
        - Tabify dock widgets for better organization.
        - Raise the first dock widget to the top.
        - Update the window title with the current project name.

        Args:
            self: The instance of the MainWindow class.
        """
        from activity_browser.layouts import panes

        # Clear all existing panes in the main window
        self.clearPanes()

        dws = []
        # Iterate through the default panes and add them as dock widgets
        for pane_class in panes.default_panes:
            pane = pane_class(parent=self)
            dockwidget = pane.getDockWidget(self)
            dws.append(dockwidget)

            # Add the dock widget to the left dock area
            self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, dockwidget)
            # Add the toggle view action to the menu bar
            self.menu_bar.view_menu.addAction(dockwidget.toggleViewAction())

            # Hide the dock widget if it is marked as hidden
            if pane_class in panes.hidden_panes:
                dockwidget.hide()

            # Synchronize the pane
            pane.sync()

        # Tabify the dock widgets for better organization
        for dw in dws:
            if dw == dws[0]:
                continue
            self.tabifyDockWidget(dws[0], dw)

        # Raise the first dock widget to the top
        dws[0].raise_()

        # Update the window title to reflect the current project
        self.setWindowTitle(f"Activity Browser - {bd.projects.current}")

    def connect_signals(self):
        # Keyboard shortcuts
        signals.project.changed.connect(self.sync)

    def clearPanes(self):
        for pane in self.panes():
            pane.deleteLater()

    def panes(self):
        """
        Return a list of all panes in the main window.
        """
        from activity_browser.ui import widgets
        return self.findChildren(widgets.ABAbstractPane)

    def set_titlebar(self):
        self.setWindowTitle(f"Activity Browser - {bd.projects.current}")

    def dialog_on_exception(self, exception: Exception):
        QtWidgets.QMessageBox.critical(
            self,
            f"An error occurred: {type(exception).__name__}",
            f"An error occurred, check the logs for more information \n\n {str(exception)}",
            QtWidgets.QMessageBox.Ok,
        )




# -*- coding: utf-8 -*-
import os
from logging import getLogger
from pathlib import Path

from peewee import SqliteDatabase, OperationalError
from qtpy import QtCore, QtWidgets

from bw2data import projects

from activity_browser.settings import ab_settings

log = getLogger(__name__)


class SettingsWizard(QtWidgets.QWizard):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.last_project = projects.current
        self.last_bwdir = projects._base_data_dir

        self.setWindowTitle("Activity Browser Settings")
        self.settings_page = SettingsPage(self)
        self.addPage(self.settings_page)
        self.show()
        self.button(QtWidgets.QWizard.BackButton).hide()
        self.button(QtWidgets.QWizard.FinishButton).clicked.connect(self.save_settings)
        self.button(QtWidgets.QWizard.CancelButton).clicked.connect(self.cancel)

    def save_settings(self):
        # directory
        current_bw_dir = ab_settings.current_bw_dir
        field = self.field("current_bw_dir")
        if field and field != current_bw_dir:
            ab_settings.custom_bw_dir = field
            ab_settings.current_bw_dir = field
            log.info(f"Saved startup brightway directory as: {field}")

        # project
        field_project = self.field("startup_project")
        current_startup_project = ab_settings.startup_project
        if field_project and field_project != current_startup_project:
            new_startup_project = field_project
            ab_settings.startup_project = new_startup_project
            log.info(f"Saved startup project as: {new_startup_project}")

        ab_settings.write_settings()
        projects.change_base_directories(Path(field))

    def cancel(self):
        log.info("Going back to before settings were changed.")
        if projects._base_data_dir != self.last_bwdir:
            projects.change_base_directories(Path(self.last_bwdir))
            projects.set_current(
                self.last_project
            )  # project changes only if directory is changed


class SettingsPage(QtWidgets.QWizardPage):
    # TODO Look to add a hover event for switching spaces
    def __init__(self, parent=None):
        super().__init__(parent)
        self.wizard = parent
        self.complete = False

        # bw dir
        self.bwdir_variables = set()
        self.bwdir = QtWidgets.QComboBox()

        self.bwdir_browse_button = QtWidgets.QPushButton("Browse")
        self.bwdir_remove_button = QtWidgets.QPushButton("Remove")
        self.update_combobox(self.bwdir, ab_settings.custom_bw_dir)
        self.restore_defaults_button = QtWidgets.QPushButton("Restore defaults")
        self.bwdir_name = QtWidgets.QLineEdit(self.bwdir.currentText())
        self.registerField("current_bw_dir", self.bwdir_name)

        # startup project
        self.startup_project_combobox = QtWidgets.QComboBox()
        self.update_project_combo()

        self.registerField(
            "startup_project", self.startup_project_combobox, "currentText"
        )

        # light/dark theme
        self.theme_combo = QtWidgets.QComboBox()
        self.theme_combo.addItems([
            "Light theme",
            "Dark theme compatibility"
        ])
        self.theme_combo.setCurrentText(ab_settings.theme)
        self.registerField(
            "theme_cbox", self.theme_combo, "currentText"
        )

        # Startup options
        self.startup_groupbox = QtWidgets.QGroupBox("Startup Options")
        self.startup_layout = QtWidgets.QGridLayout()
        self.startup_layout.addWidget(QtWidgets.QLabel("Brightway Dir: "), 0, 0)
        self.startup_layout.addWidget(self.bwdir, 0, 1)
        self.startup_layout.addWidget(self.bwdir_browse_button, 0, 2)
        self.startup_layout.addWidget(self.bwdir_remove_button, 0, 3)
        self.startup_layout.addWidget(QtWidgets.QLabel("Startup Project: "), 1, 0)
        self.startup_layout.addWidget(self.startup_project_combobox, 1, 1)
        self.startup_layout.addWidget(QtWidgets.QLabel("Theme: "), 2, 0)
        self.startup_layout.addWidget(self.theme_combo, 2, 1)
        self.startup_layout.addWidget(QtWidgets.QLabel("(Requires restart)"), 2, 2)

        self.startup_groupbox.setLayout(self.startup_layout)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.startup_groupbox)
        self.layout.addStretch()
        self.layout.addWidget(self.restore_defaults_button)
        self.setLayout(self.layout)
        self.setFinalPage(True)
        self.setButtonText(QtWidgets.QWizard.FinishButton, "Save")

        # signals
        self.startup_project_combobox.currentIndexChanged.connect(self.changed)
        self.bwdir_browse_button.clicked.connect(self.bwdir_browse)
        self.bwdir_remove_button.clicked.connect(self.bwdir_remove)
        self.bwdir.currentTextChanged.connect(self.bwdir_change)
        self.theme_combo.currentTextChanged.connect(self.theme_change)
        self.restore_defaults_button.clicked.connect(self.restore_defaults)

    def bw_projects(self, path: str):
        """Finds the bw_projects from the brightway2 environment provided by path"""
        # open the project database
        database_file = os.path.join(path, "projects.db")
        if not os.path.exists(database_file):
            return []
        db = SqliteDatabase(database_file)

        # find all project names using sql query and return
        try:
            cursor = db.execute_sql('SELECT "name" FROM "projectdataset"')
        except OperationalError as e:
            if "no such table" in str(e):
                return []
            raise
        return [i[0] for i in cursor.fetchall()]

    def restore_defaults(self):
        self.change_bw_dir(ab_settings.get_default_directory())
        self.startup_project_combobox.setCurrentText(
            ab_settings.get_default_project_name()
        )

    def bwdir_remove(self):
        """
        Removes the project from the AB settings, has additional possiblity of removing data
        contained on 'disk'. Provides a warning before execution.
        """
        hard_deletion = QtWidgets.QMessageBox.question(
            self,
            "Delete Brightway2 directory?",
            "This action will remove the local information only, click"
            "'Yes' to remove\nthe projects. Data on the \"disk\" will remain"
            " untouched and needs to be removed manually",
            QtWidgets.QMessageBox.Yes,
            QtWidgets.QMessageBox.Cancel,
        )
        if hard_deletion == QtWidgets.QMessageBox.Cancel:
            return

        removed_dir = self.bwdir.currentText()
        removed_index = self.bwdir.currentIndex()
        self.bwdir.blockSignals(True)
        self.bwdir.setCurrentIndex(-1)
        self.bwdir.removeItem(removed_index)
        self.bwdir.blockSignals(False)
        self.bwdir_variables.remove(removed_dir)
        ab_settings.remove_custom_bw_dir(removed_dir)

    def bwdir_change(self, path: str):
        """
        Executes on emission of a signal from changes to the QComboBox holding bw2 environments
        Scope: Limited to
            SettingsPage class - can create new environments and bw2data.projects (exceptions are permitted), will update
                contents of the Project QComboBox
            settings::ABSettings - uses but doesn't set bw2 variables, sets variables in the settings file
        """
        self.change_bw_dir(path)

    def theme_change(self, theme: str):
        """Change the theme."""
        if ab_settings.theme != theme:
            ab_settings.theme = theme
            self.changed()

    def bwdir_browse(self):
        """
        Executes on emission of a signal from the browse button
        Scope: Limited to
            SettingsPage class - provides a file path as a string to the QComboBox holding
                bw2data environments
        """
        path = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Select a brightway2 database folder"
        )
        if path:
            self.change_bw_dir(os.path.normpath(path))

    def change_bw_dir(self, path):
        """Set startup brightway directory.
        Switch to this directory if user wishes (this will update the "projects" combobox correctly).
        """

        # if no projects exist in this directory: ask user if he wants to set up a new brightway data directory here
        if not os.path.isfile(os.path.join(path, "projects.db")):
            create_new_directory = QtWidgets.QMessageBox.question(
                self,
                "New brightway data directory?",
                'This directory does not contain any projects. \n Would you like to setup a new brightway data directory here? \n This will close the current project and create a "default" project in the new directory.',
                QtWidgets.QMessageBox.Yes,
                QtWidgets.QMessageBox.Cancel,
            )
            if create_new_directory == QtWidgets.QMessageBox.Cancel:
                return
            else:
                self.bwdir_name.setText(path)
                self.registerField("current_bw_dir", self.bwdir_name)
                self.combobox_add_dir(self.bwdir, path)
                #                ab_settings.current_bw_dir = path
                ab_settings.startup_project = ""
                self.bwdir.blockSignals(True)
                self.bwdir.setCurrentText(self.bwdir_name.text())
                self.bwdir.blockSignals(False)
                self.update_project_combo(path=self.bwdir_name.text())
                self.changed()
        else:  # a project already exists in this directory
            # ask user if to switch directory (which will update the project combobox correctly)
            reply = QtWidgets.QMessageBox.question(
                self,
                "Continue?",
                'Would you like to switch to this directory now? \nThis will close your currently opened project. \nClick "Yes" to be able to choose the startup project.',
                QtWidgets.QMessageBox.Yes,
                QtWidgets.QMessageBox.No,
            )
            if path not in self.bwdir_variables:
                self.combobox_add_dir(self.bwdir, path)
            if reply == QtWidgets.QMessageBox.Yes:
                self.bwdir_name.setText(path)
                self.registerField("current_bw_dir", self.bwdir_name)
                #                ab_settings.current_bw_dir = path
                self.update_project_combo(path=self.bwdir_name.text())
            else:
                prev_env_index = self.bwdir.findText(
                    self.bwdir_name.text(), QtCore.Qt.MatchFixedString
                )
                self.bwdir.blockSignals(True)
                self.bwdir.setCurrentIndex(prev_env_index)
                self.bwdir.blockSignals(False)
                self.changed()

    def update_project_combo(self, initialization: bool = True, path: str = None):
        """
        Updates the project combobox when loading a new brightway environment
        """
        self.startup_project_combobox.clear()
        if path:
            self.project_names = self.bw_projects(path)
        else:
            self.project_names = self.bw_projects(ab_settings.current_bw_dir)
        if self.project_names:
            self.startup_project_combobox.addItems(self.project_names)
        else:
            log.warning("No projects found in this directory.")
        if ab_settings.startup_project in self.project_names:
            self.startup_project_combobox.setCurrentText(ab_settings.startup_project)
        else:
            ab_settings.startup_project = ""
            self.startup_project_combobox.setCurrentIndex(-1)
        if not initialization:
            self.changed()

    def combobox_add_dir(self, box: QtWidgets.QComboBox, path: str) -> None:
        """Adds a single directory to the QComboBox."""
        box.blockSignals(True)
        box.addItems([path])
        box.blockSignals(False)
        if path not in self.bwdir_variables:
            self.bwdir_variables.add(path)
            ab_settings.custom_bw_dir = path

    def update_combobox(self, box: QtWidgets.QComboBox, labels: list) -> None:
        """Update the combobox menu."""
        correct_settings = False
        current_dir = ab_settings.current_bw_dir
        for i, dir in enumerate(ab_settings.custom_bw_dir):
            self.bwdir_variables.add(dir)
            if dir == current_dir:
                box.blockSignals(True)
                box.clear()
                box.insertItems(0, labels)
                box.blockSignals(False)
                box.setCurrentIndex(i)
                correct_settings = True
        if correct_settings:
            return
        QtWidgets.QMessageBox.warning(
            self,
            "Discrepancy in the ABsettings.json file",
            "The value provided for the current brightway directory does not exist\n"
            "in the available list of directories. Please check the settings file.",
            QtWidgets.QMessageBox.Ok,
        )

    def changed(self):
        self.wizard.button(QtWidgets.QWizard.BackButton).hide()
        self.complete = True
        self.completeChanged.emit()

    def isComplete(self):
        return self.complete


import json
import os
from abc import abstractmethod
from copy import deepcopy
from typing import Type
from logging import getLogger

from qtpy import QtWebChannel, QtWebEngineWidgets, QtWidgets
from qtpy.QtCore import QObject, Qt, QUrl, Signal, Slot

from activity_browser import signals
from activity_browser.settings import ab_settings
from activity_browser.mod import bw2data as bd

from ... import utils
from ...ui.icons import qicons
from . import webutils
from .webengine_page import Page

log = getLogger(__name__)


class BaseNavigatorWidget(QtWidgets.QWidget):
    HELP_TEXT = """
    This is the text shown when the user presses 'help'.
    """
    HTML_FILE = ""

    def __init__(self, parent=None, css_file: str = "", *args, **kwargs):
        super().__init__(parent)

        # Graph object subclassed from BaseGraph.
        self.graph: Type[BaseGraph]

        # Setup JS / Qt interactions
        self.bridge = Bridge(self)
        self.channel = QtWebChannel.QWebChannel(self)
        self.channel.registerObject("bridge", self.bridge)
        self.view = QtWebEngineWidgets.QWebEngineView(self)
        self.page = Page(self.view)
        self.view.setPage(self.page)
        self.view.loadFinished.connect(self.load_finished_handler)
        self.view.setContextMenuPolicy(Qt.PreventContextMenu)
        self.view.page().setWebChannel(self.channel)
        self.url = QUrl.fromLocalFile(self.HTML_FILE)
        self.css_file = css_file

        # Various Qt objects
        self.label_help = QtWidgets.QLabel(self.HELP_TEXT)
        self.button_toggle_help = QtWidgets.QPushButton("Help")
        self.button_back = QtWidgets.QPushButton(qicons.backward, "")
        self.button_forward = QtWidgets.QPushButton(qicons.forward, "")
        self.button_refresh = QtWidgets.QPushButton("Refresh HTML")
        self.button_random_activity = QtWidgets.QPushButton("Random Activity")

    def load_finished_handler(self, *args, **kwargs) -> None:
        """Executed when webpage has been loaded for the first time or refreshed.

        Can be used to trigger a calculation after the webpage has been
        completely loaded.
        """
        pass

    @abstractmethod
    def connect_signals(self) -> None:
        self.button_toggle_help.clicked.connect(self.toggle_help)
        self.button_back.clicked.connect(self.go_back)
        self.button_forward.clicked.connect(self.go_forward)
        self.button_refresh.clicked.connect(self.draw_graph)
        self.button_random_activity.clicked.connect(self.random_graph)

    @abstractmethod
    def construct_layout(self) -> None:
        pass

    def toggle_help(self) -> None:
        self.label_help.setVisible(self.label_help.isHidden())

    def go_forward(self) -> None:
        if self.graph.forward():
            signals.new_statusbar_message.emit("Going forward.")
            self.send_json()
        else:
            signals.new_statusbar_message.emit("No data to go forward to.")

    def go_back(self) -> None:
        if self.graph.back():
            signals.new_statusbar_message.emit("Going back.")
            self.send_json()
        else:
            signals.new_statusbar_message.emit("No data to go back to.")

    def send_json(self) -> None:
        if self.graph.json_data is None:
            return
        self.bridge.graph_ready.emit(self.graph.json_data)
        css_path = webutils.get_static_css_path(self.css_file)
        css_code = utils.read_file_text(css_path)
        style_element = "<style>" + css_code + "</style>"
        self.bridge.style.emit(style_element)

    def draw_graph(self) -> None:
        self.view.load(self.url)

    @abstractmethod
    def random_graph(self) -> None:
        pass


ALL_FILTER = "All Files (*.*)"


def savefilepath(default_file_name: str, file_filter: str = ALL_FILTER):
    default = default_file_name or "Graph SVG Export"
    safe_name = bd.utils.safe_filename(default, add_hash=False)
    filepath, _ = QtWidgets.QFileDialog.getSaveFileName(
        caption="Choose location to save svg",
        dir=os.path.join(ab_settings.data_dir, safe_name),
        filter=file_filter,
    )
    return filepath


def to_svg(svg):
    """Export to .svg format."""
    # TODO: Exported filename
    filepath = savefilepath(default_file_name="svg_export", file_filter="SVG (*.svg)")
    if filepath:
        if not filepath.endswith(".svg"):
            filepath += ".svg"
        svg_file = open(filepath, "w", encoding="utf-8")
        svg_file.write(svg)
        svg_file.close()


class Bridge(QObject):
    graph_ready = Signal(str)
    update_graph = Signal(object)
    style = Signal(str)

    @Slot(str, name="node_clicked")
    def node_clicked(self, click_text: str):
        """Is called when a node is clicked in Javascript.
        Args:
            click_text: string of a serialized json dictionary describing
            - the node that was clicked on
            - mouse button and additional keys pressed
        """
        click_dict = json.loads(click_text)
        click_dict["key"] = (
            click_dict["database"],
            click_dict["id"],
        )  # since JSON does not know tuples
        log.info(f"Click information: {click_dict}")  # TODO click_dict needs correcting
        self.update_graph.emit(click_dict)

    @Slot(str, name="download_triggered")
    def download_triggered(self, svg: str):
        """Is called when a node is clicked in Javascript.
        Args:
            svg: string of svg
        """
        to_svg(svg)


class BaseGraph(object):
    def __init__(self):
        self.json_data = None
        # stores previous graphs, if any, and enables back/forward buttons
        self.stack = []
        # stores graphs that can be returned to after having used the "back" button
        self.forward_stack = []

    def update(self, delete_unstacked: bool = True) -> None:
        self.store_previous()
        if delete_unstacked:
            self.forward_stack = []

    def forward(self) -> bool:
        """Go forward, if previously gone back."""
        if not self.forward_stack:
            return False
        self.retrieve_future()
        self.update(delete_unstacked=False)
        return True

    def back(self) -> bool:
        """Go back to previous graph, if any."""
        if len(self.stack) <= 1:
            return False
        self.store_future()
        self.update(delete_unstacked=False)
        return True

    def store_previous(self) -> None:
        """Store the current graph in the"""
        self.stack.append((deepcopy(self.json_data)))

    def store_future(self) -> None:
        """When going back, store current data in a queue."""
        self.forward_stack.append(self.stack.pop())
        self.json_data = self.stack.pop()

    def retrieve_future(self) -> None:
        """Extract the last graph from the queue."""
        self.json_data = self.forward_stack.pop()

    @abstractmethod
    def new_graph(self, *args, **kwargs) -> None:
        pass

    def save_json_to_file(self, filename: str = "graph_data.json") -> None:
        """Writes the current model´s JSON representation to the specifies file."""
        if self.json_data:
            filepath = os.path.join(os.path.dirname(__file__), filename)
            with open(filepath, "w") as outfile:
                json.dump(self.json_data, outfile)



from logging import getLogger

from qtpy import QtWidgets, QtCore
from qtpy.QtCore import Signal, SignalInstance

from activity_browser import application
from activity_browser.mod import bw2data as bd
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui import icons, threading, widgets
from activity_browser.bwutils.io.ecoinvent_lcia_importer import EcoinventLCIAImporter

log = getLogger(__name__)


class MethodImporterEcoinvent(ABAction):
    """ABAction to import methods from ecoinvent"""

    icon = icons.qicons.import_db
    text = "Import methods from ecoinvent excel format"
    tool_tip = "Import methods from ecoinvent excel format"

    @classmethod
    @exception_dialogs
    def run(cls):
        # get the path from the user
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent=application.main_window,
            caption='Choose ecoinvent methods excel to import',
            filter='excel spreadsheet (*.xlsx);; All files (*.*)'
        )
        if not path:
            return

        # initialize the import thread, setting needed attributes
        extract_thread = ExtractExcelThread(application)
        extract_thread.path = path
        extract_thread.loaded.connect(cls.write_database)

        # show progress dialog for importing the excel
        progress_dialog = widgets.ABProgressDialog.get_connected_dialog("Importing Database")
        extract_thread.finished.connect(progress_dialog.deleteLater)

        extract_thread.start()

    @staticmethod
    def write_database(importer: EcoinventLCIAImporter):
        # show the import setup dialog
        import_dialog = ImportSetupDialog(importer, application.main_window)
        if import_dialog.exec_() == QtWidgets.QDialog.Rejected:
            return

        # setup the importer thread
        importer_thread = ImportExcelThread(application)
        importer_thread.importer = importer
        importer_thread.biosphere_name = import_dialog.biosphere_name
        importer_thread.prepend = import_dialog.prepend

        # setup a progress dialog
        progress_dialog = widgets.ABProgressDialog.get_connected_dialog("Importing Impact Categories")
        importer_thread.finished.connect(progress_dialog.deleteLater)

        progress_dialog.show()
        importer_thread.start()


class ImportSetupDialog(QtWidgets.QDialog):
    biosphere_name: str
    prepend: str

    def __init__(self, importer: EcoinventLCIAImporter, parent=None):
        super().__init__(parent)
        self.importer = importer

        self.setWindowTitle("Import methods from ecoinvent Excel")

        self.db_chooser = widgets.ABComboBox.get_database_combobox(self)
        self.button_comp = composites.HorizontalButtonsComposite("Cancel", "*OK")

        self.info = QtWidgets.QLabel()
        self.info.setWordWrap(True)
        self.info.setTextFormat(QtCore.Qt.RichText)

        self.prepend_label = QtWidgets.QLabel("Prepend method names")

        self.prepend_textbox = QtWidgets.QLineEdit()
        self.prepend_textbox.setPlaceholderText("Enter name prepend")
        self.prepend_textbox.textChanged.connect(self.check_overwrite)

        # Connect the necessary signals
        self.button_comp["OK"].clicked.connect(self.accept)
        self.button_comp["Cancel"].clicked.connect(self.reject)

        # Create final layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(QtWidgets.QLabel("Choose biosphere database:"))
        layout.addWidget(self.db_chooser)
        layout.addWidget(self.prepend_label)
        layout.addWidget(self.prepend_textbox)
        layout.addWidget(self.info)
        layout.addWidget(self.button_comp)

        # Set the dialog layout
        self.setLayout(layout)
        self.validate()
        self.check_overwrite()

    def check_overwrite(self, prepend=None) -> int:
        overwrite = 0
        for name in [x["name"] for x in self.importer.data]:
            if prepend:
                name = tuple([prepend, *name])

            if name in bd.methods:
                overwrite += 1

        if not overwrite:
            self.info.setText("")
            return overwrite

        self.info.setText(
            f"<p style='color: red; font-size: small;'>This action will overwrite {overwrite} impact categories</p>"
        )

        return overwrite

    def validate(self):
        """Validate the user input and enable the OK button if all is clear"""
        valid = True
        self.button_comp["OK"].setEnabled(valid)

    def accept(self):
        """Correctly set the dialog's attributes for further use in the action"""
        self.biosphere_name = self.db_chooser.currentText()
        self.prepend = self.prepend_textbox.text()
        super().accept()


class ExtractExcelThread(threading.ABThread):
    loaded: SignalInstance = Signal(EcoinventLCIAImporter)
    path: str

    def run_safely(self):
        importer = EcoinventLCIAImporter.setup_with_ei_excel(self.path)
        self.loaded.emit(importer)


class ImportExcelThread(threading.ABThread):
    biosphere_name: str
    prepend: str
    importer: EcoinventLCIAImporter

    def run_safely(self):
        self.importer.set_biosphere(self.biosphere_name)
        self.importer.apply_strategies()
        self.importer.prepend_methods(self.prepend)
        self.importer.write_methods(overwrite=True)




import pandas as pd
from qtpy import QtCore, QtWidgets

from ...ui.icons import qicons

"""
    The basic premise of this module is to contain a series of different popup menus that will allow the user
    to make a choice that can help them to resolve an issue with their use of the AB.
    
    To this end there are two small supporting classes (for a table) and the actual Popup.
    
    The first use case is with scenario files (hence the current location of this module in 
    bwutils.superstructure
    
"""


class ProblemDataModel(QtCore.QAbstractTableModel):
    """
    A simple table model for use in the ABPopup dialogs for error reporting.

    Intentionally coupled with the ABPopup class and not intended for use externally.

    """

    updated = QtCore.Signal()

    def __init__(self):
        super().__init__()
        self.columns = None
        self._dataframe = None

    def rowCount(self, *args, **kwargs):
        return self._dataframe.shape[0]

    def columnCount(self, *args, **kwargs):
        return self._dataframe.shape[1]

    def data(self, index, role=QtCore.Qt.DisplayRole):
        v = None
        if role == QtCore.Qt.DisplayRole:
            v = self._dataframe.iloc[index.row(), index.column()]
            # if the data is a tuple force the conversion to strings
            if isinstance(v, tuple):
                v = str(v)
        return v

    def sync(self, *args, **kwargs) -> None:
        assert "dataframe" in kwargs and "columns in kwargs"
        self.columns = kwargs["columns"]
        data = kwargs["dataframe"]
        self._dataframe = pd.DataFrame(data, columns=self.columns)
        self.updated.emit()

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self.columns[section]


class ProblemDataFrame(QtWidgets.QTableView):
    """
    A simple view class coupled with the ABPopup class.

    Not intended for external use.
    """

    def __init__(
        self, parent: QtWidgets.QWidget, dataframe: pd.DataFrame, cols: pd.Index
    ):
        super().__init__(parent)
        self.model = ProblemDataModel()
        self.model.updated.connect(self.update_proxy)
        self.model.sync(dataframe=dataframe, columns=cols)

    def update(self, dataframe: pd.DataFrame, cols: pd.Index):
        self.model.sync(dataframe=dataframe, columns=cols)

    def update_proxy(self):
        self.proxy = QtCore.QIdentityProxyModel(self)
        self.proxy.setSourceModel(self.model)
        self.setModel(self.proxy)


class ABPopup(QtWidgets.QDialog):
    """
    Holds AB defined message boxes to enable a more consistent popup message structure for errors.

    Primarily concerned with the creation of errors for the purposes of scenario file imports.

    Contains a tightly coupled dataframe that is intended to hold a limited sample set for user guidance.
    Dataframe management is through the dataframe() method that takes the dataframe with the data of interest
    and the columns to be extract for the creation of the internal table

    Contains the option of saving the warning output to a file with the option of printing the data that
    generated the throwing of the popup, or the full file with the error causing entries highlighted



    """

    def __init__(self):
        super().__init__()
        self.data_frame = ProblemDataFrame(self, pd.DataFrame({}), pd.Index([]))
        self.data_frame.setVisible(False)
        self._dataframe = None
        self._flags = None
        self.message = None
        self.topic = None
        self.save = None
        self.label = None
        self.buttons = QtWidgets.QHBoxLayout()
        self.button1 = None
        self.button2 = None
        self.layout = QtWidgets.QVBoxLayout()
        self.check_box = QtWidgets.QCheckBox()
        self.check_box.setVisible(False)
        self.button_layout = QtWidgets.QHBoxLayout()
        self.button_layout.addWidget(self.check_box)
        self.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum
            )
        )

    def dataframe(self, data: pd.DataFrame, columns: list = None):
        """
        Handles the creation of the internal dataframe using just those columns provided

        Arguments
        ---------
        data: a dataframe with the exchanges/rows that generate the error
        columns: a list of columns to provide the dataframe with for the popup message
        """
        dataframe = data
        cols = pd.Index(columns)
        dataframe = dataframe.loc[:, columns]
        if not isinstance(dataframe.index, pd.MultiIndex):
            dataframe.index = dataframe.index.astype(str)
        self.data_frame.update(dataframe, cols)
        self.data_frame.setHidden(False)
        self.updateGeometry()

    def save_options(self):
        """
        Creates a checkbox for determining the format for saved files
        """
        self.check_box.setVisible(True)
        self.check_box.setTristate(False)
        self.check_box.setText("Excerpt")
        self.check_box.setToolTip(
            "If left unchecked the entire file is written with an additional column indicating "
            "the status of the exchange data in the scenario file.<br> Check to save a smaller "
            "excerpt of the file, containing only those exchanges that failed."
        )
        self.check_box.setChecked(False)
        self.updateGeometry()

    def dataframe_to_file(
        self, dataframe: pd.DataFrame, flags: pd.Index = None
    ) -> None:
        """
        Sets the class variables for determining those elements of the dataframe that contain error causing data

        Arguments
        ---------
        dataframe: the pandas dataframe with the full data from importing into the AB
        flags: the pandas row index indicating those rows with the data causing the error
        """
        self._dataframe = dataframe
        self._flags = flags

    def save_dataframe(self):
        """
        Saves the dataframe according to pre-specified conditions with the dataframe_to_file and save_options class
        methods
        """
        if self.check_box.isChecked():
            self._dataframe = self._dataframe.loc[self._flags]
        elif self._dataframe is not None:
            self._dataframe["Failed"] = [False for i in self._dataframe.index]
            self._dataframe.loc[self._flags, "Failed"] = True
        # Else we're not actually intending on saving anything
        else:
            return True
        filepath, _ = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            caption="Choose the location to save the dataframe",
            filter="Excel (*.xlsx *.xls);; CSV (*.csv)",
        )
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        if filepath.endswith(".xlsx") or filepath.endswith(".xls"):
            self._dataframe.to_excel(filepath, index=False)
            QtWidgets.QApplication.restoreOverrideCursor()
            return True
        elif not filepath.endswith(".csv"):
            filepath += ".csv"
        self._dataframe.to_csv(filepath, index=False, sep=";")
        QtWidgets.QApplication.restoreOverrideCursor()
        return True

    @QtCore.Slot(name="affirmative")
    def affirmative(self):
        if self.save_dataframe():
            self.accept()

    @QtCore.Slot(name="rejection")
    def rejection(self):
        self.reject()

    # TODO set a max_width to the windows

    @staticmethod
    def abQuestion(title, message, button1, button2):
        """
        Creates an ABPopup object that contains a title message and multiple options

        Arguments
        ---------
        title: The Popup's title, should be relevant for the request to the user
        message: A detailed explanation providing why a response from the user is required, what is done
        according to the type of response and what the user should expect
        button1: a QPushButton instance MUST BE PROVIDED
        button2: a QPushButton instance MUST BE PROVIDED

        Returns
        -------
        An ABPopup instance that provides the basic format and dialog for the popup window.
        Further manipulation of the object and execution (via .exec_()) is performed upon instantiation
        """
        obj = ABPopup()
        obj.layout = QtWidgets.QVBoxLayout()
        obj.setWindowTitle(title)
        obj.label = QtWidgets.QLabel(message)
        obj.label.setWordWrap(True)
        obj.layout.addWidget(obj.label)
        obj.setWindowIcon(qicons.question)
        # add the interactive elements AND table
        obj.button1 = button1
        obj.button2 = button2
        obj.buttons.addWidget(obj.button1)
        obj.buttons.addWidget(obj.button2)
        obj.button_layout.addLayout(obj.buttons)
        obj.layout.addWidget(obj.data_frame)
        obj.layout.addWidget(obj.button_layout)

        obj.data_frame.setHidden(True)
        obj.button1.setDefault(True)
        obj.button1.clicked.connect(obj.affirmative)
        obj.button2.clicked.connect(obj.rejection)
        obj.setLayout(obj.layout)
        obj.updateGeometry()
        return obj

    @staticmethod
    def abWarning(title, message, button1, button2=None, default=1):
        """
        Creates an ABPopup object that contains a title message and a user response for a raised warning

        Arguments
        ---------
        title: The Popup's title, should be relevant for the request to the user
        message: A detailed explanation providing why a response from the user is required, what is done
        according to the type of response and what the user should expect
        button1: a QPushButton instance MUST BE PROVIDED
        button2: a QPushButton instance OPTIONAL
        default: the default button to be used (default set to button1)

        Returns
        -------
        An ABPopup instance that provides the basic format and dialog for the popup window to provide a warning.
        Further manipulation of the object and execution (via .exec_()) is performed upon instantiation
        """
        obj = ABPopup()
        obj.layout = QtWidgets.QVBoxLayout()
        obj.setWindowTitle(title)
        obj.label = QtWidgets.QLabel(message)
        obj.label.setWordWrap(True)
        obj.layout.addWidget(obj.label)
        obj.setWindowIcon(qicons.warning)

        # add the interactive elements
        obj.button1 = button1
        obj.button2 = button2
        if button2:
            obj.buttons.addWidget(obj.button1)
            obj.buttons.addWidget(obj.button2)
        else:
            obj.buttons.addWidget(button1)
        if default == 1:
            obj.button1.setDefault(True)
        else:
            obj.button2.setDefault(True)
        obj.button_layout.addLayout(obj.buttons)
        obj.layout.addWidget(obj.data_frame)
        obj.layout.addLayout(obj.button_layout)

        obj.data_frame.setHidden(True)
        obj.button1.clicked.connect(obj.affirmative)
        obj.button2.clicked.connect(obj.rejection)
        obj.setLayout(obj.layout)
        obj.updateGeometry()
        return obj

    @staticmethod
    def abCritical(title, message, button1, button2=None, default=1):
        """
        Creates an ABPopup object that contains a title message and a user response for a critical error

        Arguments
        ---------
        title: The Popup's title, should be relevant for the request to the user
        message: A detailed explanation providing why a response from the user is required, what is done
        according to the type of response and what the user should expect
        button1: a QPushButton instance MUST BE PROVIDED
        button2: a QPushButton instance OPTIONAL
        default: the default button to be used (default set to button1)

        Returns
        -------
        An ABPopup instance that provides the basic format and dialog for the popup window to provide a warning.
        Further manipulation of the object and execution (via .exec_()) is performed upon instantiation
        """
        obj = ABPopup()
        obj.layout = QtWidgets.QVBoxLayout()
        obj.setWindowTitle(title)
        obj.label = QtWidgets.QLabel(message)
        obj.label.setWordWrap(True)
        obj.layout.addWidget(obj.label)
        obj.setWindowIcon(qicons.critical)

        # add the interactive elements
        obj.button1 = button1
        obj.button2 = button2
        if button2:
            obj.buttons.addWidget(button1)
            obj.buttons.addWidget(button2)
        else:
            obj.buttons.addWidget(button1)
        if default == 1:
            obj.button1.setDefault(True)
        else:
            obj.button2.setDefault(True)
        obj.button_layout.addLayout(obj.buttons)
        obj.layout.addWidget(obj.data_frame)
        obj.layout.addLayout(obj.button_layout)

        obj.data_frame.setHidden(True)
        obj.button1.clicked.connect(obj.affirmative)
        if button2 is not None:
            obj.button2.clicked.connect(obj.rejection)
        obj.setLayout(obj.layout)
        obj.updateGeometry()
        return obj



from qtpy import QtWidgets
from qtpy.QtCore import Signal, SignalInstance


class ABFileSelector(QtWidgets.QWidget):
    textChanged: SignalInstance = Signal(str)

    def __init__(self, parent=None, filter=""):
        super().__init__(parent)

        self.filter = filter

        self.line_edit = QtWidgets.QLineEdit(self)
        self.line_edit.textChanged.connect(self.textChanged.emit)

        self.button = QtWidgets.QPushButton("Browse", self)
        self.button.clicked.connect(self.select_file)

        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(self.line_edit)
        layout.addWidget(self.button)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def select_file(self):
        file_path, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Select File", filter=self.filter)
        if file_path:
            self.line_edit.setText(file_path)

    def text(self):
        return self.line_edit.text()

    def setText(self, text):
        self.line_edit.setText(text)



from typing import Any

from activity_browser.actions.base import ABAction, exception_dialogs
from bw2data.parameters import (ActivityParameter, Group,
                                                     GroupDependency,
                                                     parameters)
from activity_browser.ui.icons import qicons


class ParameterClearBroken(ABAction):
    """
    Take the given information and attempt to remove all the downstream parameter information.
    """

    icon = qicons.delete
    text = "Clear broken parameter"

    @staticmethod
    @exception_dialogs
    def run(parameter: Any):
        db = parameter.database
        code = parameter.code
        group = parameter.group

        # I'm not sure this is right, because you're removing all the exchanges from the group...
        parameters.remove_exchanges_from_group(group, None, False)
        ActivityParameter.delete().where(
            (ActivityParameter.database == db) & (ActivityParameter.code == code)
        ).execute()

        # Also clear Group if it is not in use anymore
        if (
            not ActivityParameter.select()
            .where(ActivityParameter.group == parameter.group)
            .exists()
        ):
            Group.delete().where(Group.name == group).execute()
            GroupDependency.delete().where(GroupDependency.group == group).execute()



from qtpy import QtWidgets

import bw2data as bd
import pandas as pd

from activity_browser import actions
from activity_browser.ui import widgets, delegates


class ImpactCategorySection(QtWidgets.QWidget):
    def __init__(self, calculation_setup_name: str, parent=None):
        super().__init__(parent)

        self.calculation_setup_name = calculation_setup_name
        self.calculation_setup = bd.calculation_setups.get(self.calculation_setup_name)

        self.view = ImpactCategoryView()
        self.model = ImpactCategoryModel()
        self.view.setModel(self.model)

        self.build_layout()

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        self.setLayout(layout)

    def sync(self):
        try:
            self.calculation_setup = bd.calculation_setups[self.calculation_setup_name]
            self.model.setDataFrame(self.build_df())
        except KeyError:
            self.parent().close()
            self.parent().deleteLater()

    def build_df(self):
        data = [bd.methods.get(method_name) for method_name in self.calculation_setup.get("ia", [])]
        df = pd.DataFrame(data, columns=["name", "unit", "num_cfs"])

        df["name"] = self.calculation_setup.get("ia", [])

        cols = ["name", "unit", "num_cfs"]

        return df[cols]


class ImpactCategoryView(widgets.ABTreeView):
    defaultColumnDelegates = {
        "name": delegates.StringDelegate
    }

    class ContextMenu(QtWidgets.QMenu):
        def __init__(self, pos, view: "ImpactCategoryView"):
            super().__init__(view)
            cs_name = view.parent().calculation_setup_name

            if not view.selectedIndexes():
                return

            indices = [index.internalPointer().key() for index in view.selectedIndexes()]

            self.delete_ic_action = actions.CSDeleteImpactCategory.get_QAction(cs_name, indices)
            print(self.delete_ic_action.text())
            self.addAction(self.delete_ic_action)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

    def dragMoveEvent(self, event) -> None:
        pass

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("application/bw-methodnamelist"):
            event.accept()

    def dropEvent(self, event) -> None:
        event.accept()
        cs_name = self.parent().calculation_setup_name
        method_names = event.mimeData().retrievePickleData("application/bw-methodnamelist")
        actions.CSAddImpactCategory.run(cs_name, method_names)


class ImpactCategoryItem(widgets.ABDataItem):
    pass


class ImpactCategoryModel(widgets.ABItemModel):
    dataItemClass = ImpactCategoryItem



from qtpy import QtWidgets, QtGui
from qtpy.QtCore import Qt

import bw2data as bd
import pandas as pd

from activity_browser import actions, signals
from activity_browser.ui import widgets, icons, delegates
from activity_browser.bwutils import AB_metadata


class ImpactCategoryDetailsPage(QtWidgets.QWidget):
    def __init__(self, name: tuple, parent=None):
        super().__init__(parent)
        self.name = name
        self.impact_category = bd.Method(name)

        self.setObjectName(" | ".join(name))

        self.model = CharacterizationFactorsModel(self, self.build_df())
        self.view = CharacterizationFactorsView(self)
        self.view.setModel(self.model)

        # resizing name and categories columns
        self.view.resizeColumnToContents(0)
        self.view.resizeColumnToContents(1)

        self.build_layout()
        self.connect_signals()

    def connect_signals(self):
        signals.method.deleted.connect(self.on_method_deleted)
        signals.meta.methods_changed.connect(self.sync)

    def on_method_deleted(self, method):
        if method.name == self.name:
            self.deleteLater()

    def sync(self):
        self.impact_category = bd.Method(self.name)
        self.model.setDataFrame(self.build_df())

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(widgets.ABLabel.demiBold("Impact Category: " + " - ".join(self.name), self))
        layout.addWidget(widgets.ABHLine(self))
        layout.addWidget(self.view)
        self.setLayout(layout)

    def build_df(self):
        df = pd.DataFrame(self.impact_category.load(), columns=["id", "amount"])
        other = AB_metadata.dataframe[["id", "name", "categories", "database", "unit"]]

        df = df.merge(other, left_on="id", right_on="id").rename(columns={"id": "_id"})
        df["_impact_category_name"] = [self.name for i in range(len(df))]

        cols = ["name", "categories", "database", "amount", "unit", "_id", "_impact_category_name"]
        return df[cols]


class CharacterizationFactorsView(widgets.ABTreeView):
    defaultColumnDelegates = {
        "amount": delegates.FloatDelegate,
        "categories": delegates.ListDelegate,
    }


class ExchangesItem(widgets.ABDataItem):
    def flags(self, col: int, key: str):
        """
        Returns the item flags for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the flags.

        Returns:
            QtCore.Qt.ItemFlags: The item flags.
        """
        flags = super().flags(col, key)
        if key in ["amount"]:
            return flags | Qt.ItemFlag.ItemIsEditable
        return flags

    def decorationData(self, col, key):
        """
        Provides decoration data for the item.

        Args:
            col: The column index.
            key: The key for which to provide decoration data.

        Returns:
            The decoration data for the item.
        """
        if key == "name":
            return icons.qicons.biosphere

    def fontData(self, col: int, key: str):
        """
        Returns the font data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the font data.

        Returns:
            QtGui.QFont: The font data.
        """
        font = super().fontData(col, key)

        # set the font to bold if it's a production/functional exchange
        if key == "name":
            font.setWeight(QtGui.QFont.Weight.DemiBold)
        return font

    def setData(self, col: int, key: str, value) -> bool:
        """
        Sets the data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to set the data.
            value: The value to set.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if key not in ["amount"]:
            return False

        actions.CFAmountModify.run(self["_impact_category_name"], self["_id"], value)


class CharacterizationFactorsModel(widgets.ABItemModel):
    dataItemClass = ExchangesItem



from logging import getLogger

from qtpy import QtWidgets

from activity_browser import application, signals
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class CSDelete(ABAction):
    """
    ABAction to delete a calculation setup. First asks the user for confirmation and returns if cancelled. Otherwise,
    passes the csname to the CalculationSetupController for deletion. Finally, displays confirmation that it succeeded.
    """

    icon = qicons.delete
    text = "Delete"

    @staticmethod
    @exception_dialogs
    def run(cs_names: str | list[str]):
        if isinstance(cs_names, str):
            cs_names = [cs_names]

        # ask the user whether they are sure to delete the calculation setup
        warning = QtWidgets.QMessageBox.warning(
            application.main_window,
            f"Deleting Calculation Setup(s): {', '.join(cs_names)}",
            "Are you sure?",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No,
        )

        # return if the users cancels
        if warning == QtWidgets.QMessageBox.No:
            return

        for cs_name in cs_names:
            if cs_name not in bd.calculation_setups:
                log.warning(f"Calculation setup {cs_name} not found")
                continue

            del bd.calculation_setups[cs_name]
            log.info(f"Deleted calculation setup: {cs_name}")



# -*- coding: utf-8 -*-
import json
import os
import time
from typing import List
from logging import getLogger

import bw2calc as bc
import bw2data as bd
import numpy
from bw_graph_tools.graph_traversal import Edge as GraphEdge
from bw_graph_tools.graph_traversal import NewNodeEachVisitGraphTraversal
from bw_graph_tools.graph_traversal import Node as GraphNode
from qtpy import QtWidgets
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QComboBox

from activity_browser import signals
from activity_browser.mod import bw2data as bd
from bw2data.backends import ActivityDataset

from ...bwutils.commontasks import identify_activity_type
from .base import BaseGraph, BaseNavigatorWidget

log = getLogger(__name__)


# TODO:
# switch between percent and absolute values
# when avoided impacts, then the scaling between 0-1 of relative impacts does not work properly
# ability to navigate to activities
# ability to calculate LCA for selected activities
# ability to expand (or reduce) the graph
# save graph as image
# random_graph should not work for biosphere

# in Javascript:
# - zoom behaviour


class SankeyNavigatorWidget(BaseNavigatorWidget):
    HELP_TEXT = """
    LCA Sankey:

    Red flows: Impacts
    Green flows: Avoided impacts

    """
    HTML_FILE = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../../static/sankey_navigator.html"
    )

    def __init__(self, cs_name, parent=None):
        super().__init__(parent, css_file="sankey_navigator.css")

        self.cache = {}  # we cache the calculated data to improve responsiveness
        self.parent = parent
        self.has_scenarios = self.parent.has_scenarios
        self.cs = cs_name
        self.selected_db = None
        self.has_sankey = False
        self.func_units = []
        self.methods = []
        self.scenarios = []
        self.graph = Graph()

        # Additional Qt objects
        self.scenario_label = QtWidgets.QLabel("Scenario: ")
        self.func_unit_cb = QtWidgets.QComboBox()
        self.method_cb = QtWidgets.QComboBox()
        self.scenario_cb = QtWidgets.QComboBox()
        self.cutoff_sb = QtWidgets.QDoubleSpinBox()
        self.max_calc_sb = QtWidgets.QDoubleSpinBox()
        self.button_calculate = QtWidgets.QPushButton("Calculate")
        self.layout = QtWidgets.QVBoxLayout()

        # graph
        self.draw_graph()
        self.construct_layout()
        self.connect_signals()

    @Slot(name="loadFinishedHandler")
    def load_finished_handler(self) -> None:
        if self.has_sankey:
            self.send_json()

    def connect_signals(self):
        super().connect_signals()
        self.button_calculate.clicked.connect(self.new_sankey)
        signals.database_selected.connect(self.set_database)
        # checkboxes
        self.func_unit_cb.currentIndexChanged.connect(self.new_sankey)
        self.method_cb.currentIndexChanged.connect(self.new_sankey)
        self.scenario_cb.currentIndexChanged.connect(self.new_sankey)

    def construct_layout(self) -> None:
        """Layout of Sankey Navigator"""
        super().construct_layout()
        self.label_help.setVisible(False)

        # Layout Reference Flows and Impact Categories
        grid_lay = QtWidgets.QGridLayout()
        grid_lay.addWidget(QtWidgets.QLabel("Reference flow: "), 0, 0)

        grid_lay.addWidget(self.scenario_label, 1, 0)
        grid_lay.addWidget(QtWidgets.QLabel("Impact indicator: "), 2, 0)

        self.update_calculation_setup()

        grid_lay.addWidget(self.func_unit_cb, 0, 1)
        grid_lay.addWidget(self.scenario_cb, 1, 1)
        grid_lay.addWidget(self.method_cb, 2, 1)

        # cut-off
        grid_lay.addWidget(QtWidgets.QLabel("Cutoff: "), 2, 2)
        self.cutoff_sb.setRange(0.0, 1.0)
        self.cutoff_sb.setSingleStep(0.01)
        self.cutoff_sb.setDecimals(3)
        self.cutoff_sb.setValue(0.05)
        self.cutoff_sb.setKeyboardTracking(False)
        grid_lay.addWidget(self.cutoff_sb, 2, 3)

        # max-iterations of graph traversal
        grid_lay.addWidget(QtWidgets.QLabel("Calculation depth: "), 2, 4)
        self.max_calc_sb.setRange(1, 2000)
        self.max_calc_sb.setSingleStep(50)
        self.max_calc_sb.setDecimals(0)
        self.max_calc_sb.setValue(250)
        self.max_calc_sb.setKeyboardTracking(False)
        grid_lay.addWidget(self.max_calc_sb, 2, 5)

        grid_lay.setColumnStretch(6, 1)
        hlay = QtWidgets.QHBoxLayout()
        hlay.addLayout(grid_lay)

        # Controls Layout
        hl_controls = QtWidgets.QHBoxLayout()
        hl_controls.addWidget(self.button_back)
        hl_controls.addWidget(self.button_forward)
        hl_controls.addWidget(self.button_calculate)
        hl_controls.addWidget(self.button_refresh)
        hl_controls.addWidget(self.button_random_activity)
        hl_controls.addWidget(self.button_toggle_help)
        hl_controls.addStretch(1)

        # Layout
        self.layout.addLayout(hl_controls)
        self.layout.addLayout(hlay)
        self.layout.addWidget(self.label_help)
        self.layout.addWidget(self.view)
        self.setLayout(self.layout)

    def get_scenario_labels(self) -> List[str]:
        """Get scenario labels if scenario is used."""
        return self.parent.mlca.scenario_names if self.has_scenarios else []

    def configure_scenario(self):
        """Determine if scenario Qt widgets are visible or not and retrieve
        scenario labels for the selection drop-down box.
        """
        self.scenario_cb.setVisible(self.has_scenarios)
        self.scenario_label.setVisible(self.has_scenarios)
        if self.has_scenarios:
            self.scenarios = self.get_scenario_labels()
            self.update_combobox(self.scenario_cb, self.scenarios)

    @staticmethod
    def update_combobox(box: QComboBox, labels: List[str]) -> None:
        """Update the combobox menu."""
        box.blockSignals(True)
        box.clear()
        box.insertItems(0, labels)
        box.blockSignals(False)

    def update_calculation_setup(self, cs_name=None) -> None:
        """Update Calculation Setup, reference flows and impact categories, and dropdown menus."""
        # block signals
        self.func_unit_cb.blockSignals(True)
        self.method_cb.blockSignals(True)

        self.cs = cs_name or self.cs
        self.func_units = [
            {bd.get_activity(k): v for k, v in fu.items()}
            for fu in bd.calculation_setups[self.cs]["inv"]
        ]
        self.methods = bd.calculation_setups[self.cs]["ia"]
        self.func_unit_cb.clear()
        fu_acts = [list(fu.keys())[0] for fu in self.func_units]
        self.func_unit_cb.addItems(
            [f"{repr(a)} | {a._data.get('database')}" for a in fu_acts]
        )
        self.configure_scenario()
        self.method_cb.clear()
        self.method_cb.addItems([repr(m) for m in self.methods])

        # unblock signals
        self.func_unit_cb.blockSignals(False)
        self.method_cb.blockSignals(False)

    def new_sankey(self) -> None:
        """(re)-generate the sankey diagram."""
        demand_index = self.func_unit_cb.currentIndex()
        method_index = self.method_cb.currentIndex()

        demand = self.func_units[demand_index]
        method = self.methods[method_index]
        scenario_index = None
        scenario_lca = False
        if self.has_scenarios:
            scenario_lca = True
            scenario_index = self.scenario_cb.currentIndex()
        self.update_sankey(
            demand,
            method,
            demand_index=demand_index,
            method_index=method_index,
            scenario_index=scenario_index,
            scenario_lca=scenario_lca,
            cut_off=self.cutoff_sb.value(),
            max_calc=int(self.max_calc_sb.value()),
        )

    def update_sankey(
        self,
        demand: dict,
        method: tuple,
        demand_index: int = None,
        method_index: int = None,
        scenario_index: int = None,
        scenario_lca: bool = False,
        cut_off=0.05,
        max_calc=100,
    ) -> None:
        """Calculate LCA, do graph traversal, get JSON graph data for this, and send to javascript."""

        # the cache key consists of demand/method/scenario indices (index of item in the relevant tables),
        # the cutoff, max_calc.
        # together, these are unique.
        cache_key = (demand_index, method_index, scenario_index, cut_off, max_calc)
        if data := self.cache.get(cache_key, False):
            # this Sankey is already cached, generate the Sankey with the cached data
            log.debug(f"CACHED sankey for: {demand}, {method}, key: {cache_key}")
            self.graph.new_graph(data)
            self.has_sankey = bool(self.graph.json_data)
            self.send_json()
            return

        start = time.time()
        log.debug(f"CALCULATE sankey for: {demand}, {method}, key: {cache_key}")
        try:
            if scenario_lca:
                self.parent.mlca.update_lca_calculation_for_sankey(
                    scenario_index, demand, method_index
                )
                lca = self.parent.mlca.lca
                data = NewNodeEachVisitGraphTraversal.calculate(
                    lca, cutoff=cut_off, max_calc=int(max_calc)
                )
            else:
                fu, data_objs, _ = bd.prepare_lca_inputs(demand=demand, method=method)
                lca = bc.LCA(demand=fu, data_objs=data_objs)
                lca.lci(factorize=True)
                lca.lcia()
                data = NewNodeEachVisitGraphTraversal.calculate(
                    lca_object=lca, cutoff=cut_off, max_calc=int(max_calc)
                )

            # store the metadata from this calculation
            data["metadata"] = {
                "lca": lca,
                "unit": bd.methods[method]["unit"],
            }
        except (ValueError, ZeroDivisionError) as e:
            QtWidgets.QMessageBox.information(
                None, "Nonsensical numeric result.", str(e)
            )
        log.debug(f"Completed graph traversal ({round(time.time() - start, 2)} seconds")

        # cache the generated Sankey data
        self.cache[cache_key] = data

        # generate the new Sankey
        self.graph.new_graph(data)
        self.has_sankey = bool(self.graph.json_data)
        self.send_json()

    def set_database(self, name):
        """Saves the currently selected database for graphing a random activity"""
        self.selected_db = name

    def random_graph(self) -> None:
        """Show graph for a random activity in the currently loaded database."""
        if self.selected_db:
            method = bd.methods.random()
            act = bd.Database(self.selected_db).random()
            demand = {act: 1.0}
            self.update_sankey(demand, method)
        else:
            QtWidgets.QMessageBox.information(
                None, "Not possible.", "Please load a database first."
            )


def convert_numpy_types(obj) -> int | float | list:
    """Converts numpy types into serializable types"""
    if isinstance(obj, numpy.integer):
        return int(obj)
    if isinstance(obj, numpy.floating):
        return float(obj)
    if isinstance(obj, numpy.ndarray):
        return obj.tolist()
    return obj


def make_serializable(data: dict) -> dict:
    """Converts numpy data into serializable values for json.dumps"""
    for key, value in data.items():
        if isinstance(value, dict):
            make_serializable(value)
        elif isinstance(value, list):
            data[key] = [
                (
                    convert_numpy_types(v)
                    if not isinstance(v, dict)
                    else make_serializable(v)
                )
                for v in value
            ]
        else:
            data[key] = convert_numpy_types(value)
    return data


class Graph(BaseGraph):
    """
    Python side representation of the graph.
    Functionality for graph navigation (e.g. adding and removing nodes).
    A JSON representation of the graph (edges and nodes) enables its use in javascript/html/css.
    """

    def new_graph(self, data):
        self.json_data = Graph.get_json_data(data)
        self.update()

    @staticmethod
    def get_json_data(data) -> str:
        """Transform graph traversal output to JSON data.

        We use the [dagre](https://github.com/dagrejs/dagre) javascript library for rendering directed graphs. We need to provide the following:

        ```python
        {
            'max_impact': float,  # Total LCA score,
            'title': str,  # Graph title
            'edges': [{
                'source_id': int,  # Unique ID of producer of material or energy in graph
                'target_id': int,  # Unique ID of consumer of material or energy in graph
                'weight': float,  #  In graph units, relative to `max_edge_width`
                'label': str,  # HTML label
                'product': str,  # The label of the flowing material or energy
                'class': str,  # "benefit" or "impact"; controls styling
                'label': str,  # HTML label
                'toottip': str,  # HTML tooltip
            }],
            'nodes': [{
                'direct_emissions_score_normalized': float,  # Fraction of total LCA score from direct emissions
                'product': str,  # Reference product label, if any
                'location': str,  # Location, if any
                'id': int,  # Graph traversal ID
                'database_id': int,  # Node ID in SQLite database
                'database': str,  # Database name
                'class': str,  # Enumerated set of class label strings
                'label': str,  # HTML label including name and location
                'toottip': str,  # HTML tooltip
            }],
        }

        ```

        """
        lca_score = data["metadata"]["lca"].score
        lcia_unit = data["metadata"]["unit"]
        demand = data["metadata"]["lca"].demand

        def convert_edge_to_json(
            edge: GraphEdge,
            nodes: dict[int, GraphNode],
            total_score: float,
            lcia_unit: str,
            max_edge_width: int = 40,
        ) -> dict:
            cum_score = nodes[edge.producer_unique_id].cumulative_score
            unit = bd.get_node(
                id=nodes[edge.producer_unique_id].reference_product_datapackage_id
            ).get("unit", "(unknown)")
            return {
                "source_id": edge.producer_unique_id,
                "target_id": edge.consumer_unique_id,
                "amount": edge.amount,
                "weight": abs(cum_score / total_score) * max_edge_width,
                "label": f"{round(cum_score, 3)} {lcia_unit}",
                "class": "benefit" if cum_score < 0 else "impact",
                "tooltip": f"<b>{round(cum_score, 3)} {lcia_unit}</b> ({edge.amount:.2g} {unit})",
            }

        def convert_node_to_json(
            graph_node: GraphNode,
            total_score: float,
            fu: dict,
            lcia_unit: str,
            max_name_length: int = 20,
        ) -> dict:
            db_node = bd.get_node(id=graph_node.activity_datapackage_id)
            data = {
                "direct_emissions_score_normalized": graph_node.direct_emissions_score
                / (total_score or 1),
                "direct_emissions_score": graph_node.direct_emissions_score,
                "cumulative_score": graph_node.cumulative_score,
                "cumulative_score_normalized": graph_node.cumulative_score
                / (total_score or 1),
                "product": db_node.get("reference product", ""),
                "location": db_node.get("location", "(unknown)"),
                "id": graph_node.unique_id,
                "database_id": graph_node.activity_datapackage_id,
                "database": db_node["database"],
                "class": (
                    "demand"
                    if graph_node.activity_datapackage_id in fu
                    else identify_activity_type(db_node)
                ),
                "name": db_node.get("name", "(unnamed)"),
            }
            frac_dir_score = round(data["direct_emissions_score_normalized"] * 100, 2)
            dir_score = round(data["direct_emissions_score"], 3)
            frac_cum_score = round(data["cumulative_score_normalized"] * 100, 2)
            cum_score = round(data["cumulative_score"], 3)
            data[
                "label"
            ] = f"""{db_node['name'][:max_name_length]}
{data['location']}
{frac_dir_score}%"""
            data[
                "tooltip"
            ] = f"""
                <b>{data['name']}</b>
                <br>Individual impact: {dir_score} {lcia_unit} ({frac_dir_score }%)
                <br>Cumulative impact: {cum_score} {lcia_unit} ({frac_cum_score}%)
            """
            return data

        json_data = {
            "nodes": [
                convert_node_to_json(node, lca_score, demand, lcia_unit)
                for idx, node in data["nodes"].items()
                if idx != -1
            ],
            "edges": [
                convert_edge_to_json(edge, data["nodes"], lca_score, lcia_unit)
                for edge in data["edges"]
                if edge.producer_index != -1 and edge.consumer_index != -1
            ],
            "title": "Sankey graph result",
            # "title": self.build_title(demand, lca_score, lcia_unit),
        }

        return json.dumps(json_data)

    def build_title(self, demand: tuple, lca_score: float, lcia_unit: str) -> str:
        act, amount = demand[0], demand[1]
        if type(act) is tuple or type(act) is int:
            act = bd.get_activity(act)
        format_str = (
            "Reference flow: {:.2g} {} {} | {} | {} <br>" "Total impact: {:.2g} {}"
        )
        return format_str.format(
            amount,
            act.get("unit"),
            act.get("reference product") or act.get("name"),
            act.get("name"),
            act.get("location"),
            lca_score,
            lcia_unit,
        )


def id_to_key(id):
    if isinstance(id, tuple):
        return id
    return ActivityDataset.get_by_id(id).database, ActivityDataset.get_by_id(id).code



from logging import getLogger

from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt

import pandas as pd
import bw2data as bd
import bw_functional as bf

from activity_browser import actions, bwutils
from activity_browser.bwutils import refresh_node, AB_metadata, database_is_locked, database_is_legacy
from activity_browser.ui import widgets, icons, delegates

log = getLogger(__name__)

EXCHANGE_MAP = {
    "natural resource": "biosphere", "emission": "biosphere", "inventory indicator": "biosphere",
    "economic": "biosphere", "social": "biosphere", "product": "technosphere",
    "processwithreferenceproduct": "technosphere", "waste": "technosphere",
}


class ExchangesTab(QtWidgets.QWidget):
    """
    A widget that displays exchanges related to a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display exchanges for.
        output_view (ExchangesView): The view displaying the output exchanges.
        output_model (ExchangesModel): The model containing the data for the output exchanges.
        input_view (ExchangesView): The view displaying the input exchanges.
        input_model (ExchangesModel): The model containing the data for the input exchanges.
    """
    def __init__(self, activity: tuple | int | bd.Node, parent=None):
        """
        Initializes the ExchangesTab widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display exchanges for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self.setAcceptDrops(True)

        # Refresh the activity node
        self.activity = refresh_node(activity)

        # Output Table
        self.output_view = ExchangesView(self)
        self.output_model = ExchangesModel(self)
        self.output_view.setModel(self.output_model)

        # Set indentation for output view
        self.output_view.setIndentation(0)

        # Input Table
        self.input_view = ExchangesView(self)
        self.input_model = ExchangesModel(self)
        self.input_view.setModel(self.input_model)

        # Set indentation for input view
        self.input_view.setIndentation(0)

        # Overlay for drag and drop
        self.overlay = None

        # Build the layout of the widget
        self.build_layout()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 10, 0, 1)

        # Add output label and view to the layout
        layout.addWidget(widgets.ABLabel.demiBold(" Output:", self))
        layout.addWidget(self.output_view)

        # Add input label and view to the layout
        layout.addWidget(widgets.ABLabel.demiBold(" Input:", self))
        layout.addWidget(self.input_view)

        # Set the layout for the widget
        self.setLayout(layout)

    def sync(self) -> None:
        """
        Synchronizes the widget with the current state of the activity.
        """
        # Refresh the activity node
        self.activity = refresh_node(self.activity)

        # Get the production, technosphere, and biosphere exchanges
        production = self.activity.production()
        technosphere = self.activity.technosphere()
        biosphere = self.activity.biosphere()

        # Filter inputs and outputs based on the amount and type
        inputs = ([x for x in production if x["amount"] < 0] +
                  [x for x in technosphere if x["amount"] >= 0] +
                  [x for x in biosphere if (x.input["type"] != "emission" and x["amount"] >= 0) or (x.input["type"] == "emission" and x["amount"] < 0)])

        outputs = ([x for x in production if x["amount"] >= 0] +
                   [x for x in technosphere if x["amount"] < 0] +
                   [x for x in biosphere if (x.input["type"] == "emission" and x["amount"] >= 0) or (x.input["type"] != "emission" and x["amount"] < 0)])

        # Update the models with the new data
        output_df = self.build_df(outputs)
        self.output_model.setDataFrame(output_df)
        self.output_view.drag_drop_hint.setVisible(output_df.empty)

        input_df = self.build_df(inputs)
        self.input_model.setDataFrame(input_df)
        self.input_view.drag_drop_hint.setVisible(input_df.empty)

    def build_df(self, exchanges) -> pd.DataFrame:
        """
        Builds a DataFrame from the given exchanges.

        Args:
            exchanges (list): The list of exchanges to build the DataFrame from.

        Returns:
            pd.DataFrame: The DataFrame containing the exchanges data.
        """
        # Define the columns for the metadata
        cols = ["key", "unit", "name", "product", "location", "database", "substitute", "substitution_factor", "allocation_factor",
                "properties", "processor", "categories"]

        # Create a DataFrame from the exchanges
        exc_df = pd.DataFrame(exchanges, columns=["amount", "input", "formula", "uncertainty type", "comment"])
        act_df = AB_metadata.get_metadata(exc_df["input"].unique(), cols)

        # Merge the exchanges DataFrame with the metadata DataFrame
        df = exc_df.merge(
            act_df,
            left_on="input",
            right_on="key"
        ).drop(columns=["key"])

        # Use "product" if available otherwise use "name"
        df.update(df["product"].rename("name"))

        # Handle substitute data if available
        if not df["substitute"].isna().all():
            df = df.merge(
                AB_metadata.dataframe[["key", "name"]].rename({"name": "substitute_name"}, axis="columns"),
                left_on="substitute",
                right_on="key",
                how="left",
            ).drop(columns=["key"])
        else:
            df.drop(columns=["substitute", "substitution_factor"], inplace=True)

        # Handle properties data if available
        if not act_df.properties.isna().all():
            props_df = act_df[act_df.properties.notna()]
            props_df = pd.DataFrame(list(props_df.get("properties")), index=props_df.key)
            props_df.rename(lambda col: f"property_{col}", axis="columns", inplace=True)

            df = df.merge(
                props_df,
                left_on="input",
                right_index=True,
                how="left",
            )

        # Add allocation and activity type information
        df["_allocate_by"] = self.activity.get("allocation")
        df["_activity_type"] = self.activity.get("type")
        df["_exchange"] = exchanges

        # Drop the properties column and rename some columns
        df.drop(columns=["properties"], inplace=True)
        df.rename({"input": "_input_key", "substitute": "_substitute_key", "processor": "_processor_key",
                   "uncertainty type": "uncertainty"},
            axis="columns", inplace=True)

        # Define the order of columns for the final DataFrame
        cols = ["amount", "unit", "name", "location", "categories", "database"]
        cols += ["substitute_name", "substitution_factor"] if "substitute_name" in df.columns else []
        cols += ["allocation_factor"] if not database_is_legacy(self.activity.get("database")) else []
        cols += [col for col in df.columns if col.startswith("property")]
        cols += ["formula", "comment", "uncertainty"]
        cols += [col for col in df.columns if col.startswith("_")]

        return df[cols]

    def dragEnterEvent(self, event):
        """
        Handles the drag enter event.

        Args:
            event: The drag enter event.
        """
        if database_is_locked(self.activity["database"]):
            return

        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            self.overlay = DropOverlay(self)
            self.overlay.show()
            event.accept()

    def dragLeaveEvent(self, event):
        """
        Handles the drag leave event.

        Args:
            event: The drag leave event.
        """
        # Reset the palette on drag leave
        self.overlay.deleteLater()

    def dropEvent(self, event):
        """
        Handles the drop event.

        Args:
            event: The drop event.
        """
        log.debug(f"Dropevent from: {type(event.source()).__name__} to: {self.__class__.__name__}")
        # Reset the palette on drop
        self.overlay.deleteLater()

        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
        exchanges = {"technosphere": set(), "biosphere": set()}

        for key in keys:
            act = bd.get_node(key=key)
            if act["type"] not in EXCHANGE_MAP:
                continue
            exc_type = EXCHANGE_MAP[act["type"]]
            exchanges[exc_type].add(act.key)

        # Run the action for new exchanges
        for exc_type, keys in exchanges.items():
            actions.ExchangeNew.run(keys, self.activity.key, exc_type)


class DropOverlay(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WA_TransparentForMouseEvents)
        self.setAttribute(Qt.WA_NoSystemBackground)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setAutoFillBackground(False)
        self.resize(parent.size())

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        painter.fillRect(self.rect(), QtGui.QColor(0, 100, 255, 200))  # Semi-transparent blue
        painter.setPen(Qt.white)

        font = self.font()
        font.setBold(True)

        painter.setFont(font)
        painter.drawText(self.rect(), Qt.AlignCenter, "Drop here to create new exchanges")


class ExchangesView(widgets.ABTreeView):
    """
    A view that displays the exchanges in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
        hovered_item (ExchangesItem): The item currently being hovered over.
    """
    defaultColumnDelegates = {
        "amount": delegates.AmountDelegate,
        "allocation_factor": delegates.FloatDelegate,
        "substitution_factor": delegates.FloatDelegate,
        "unit": delegates.StringDelegate,
        "name": delegates.StringDelegate,
        "location": delegates.StringDelegate,
        "product": delegates.StringDelegate,
        "formula": delegates.NewFormulaDelegate,
        "comment": delegates.StringDelegate,
        "uncertainty": delegates.UncertaintyDelegate,
        "categories": delegates.ListDelegate,
    }

    class HeaderMenu(widgets.ABMenu):
        menuSetup = [
            lambda m: m.setup_view_menu(),
            lambda m: m.setup_allocation(),
        ]

        def setup_view_menu(self):
            table_view: ExchangesView = self.parent()
            table_model: ExchangesModel = table_view.model()

            def toggle_slot(action: QtWidgets.QAction):
                """
                Toggles the visibility of columns based on the action triggered.

                Args:
                    action (QtWidgets.QAction): The action triggered.
                """
                indices = action.data()
                for index in indices:
                    hidden = table_view.isColumnHidden(index)
                    table_view.setColumnHidden(index, not hidden)

            # Create the view menu
            view_menu = QtWidgets.QMenu(table_view)
            view_menu.setTitle("View")

            props_indices = []

            # Add actions for each column to the view menu
            for i, col in enumerate(table_model.columns()):
                if col.startswith("property"):
                    props_indices.append(i)
                    continue

                action = QtWidgets.QAction(table_model.columns()[i], self)
                action.setCheckable(True)
                action.setChecked(not table_view.isColumnHidden(i))
                action.setData([i])

                view_menu.addAction(action)

            # Add a combined action for property columns
            if props_indices:
                action = QtWidgets.QAction("properties", self)
                action.setCheckable(True)
                action.setChecked(not table_view.isColumnHidden(props_indices[0]))
                action.setData(props_indices)
                view_menu.addAction(action)

            # Connect the view menu actions to the toggle slot
            view_menu.triggered.connect(toggle_slot)

            # Add the view menu to the context menu
            self.addMenu(view_menu)

        def setup_allocation(self):
            table_view: ExchangesView = self.parent()

            if database_is_locked(table_view.activity["database"]) or not self.column.startswith("property"):
                return

            action = actions.ActivityModify.get_QAction(table_view.activity.key,
                                                        "allocation",
                                                        self.column[9:],
                                                        parent=self)
            action.setText(f"Allocate by {self.column[9:]}")
            self.addAction(action)

        @property
        def column(self):
            view, model, pos = self.parent(), self.parent().model(), QtGui.QCursor.pos()
            col_index = view.columnAt(view.mapFromGlobal(pos).x())
            return model.columns()[col_index]

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m: m.add(actions.ActivityNewProduct, [m.activity.key],
                            enable=not m.locked and not database_is_legacy(m.activity["database"])
                            ),
            lambda m: m.add(actions.ExchangeDelete, m.exchanges, enable=bool(m.exchanges) and not m.locked),
            lambda m: m.add(actions.ExchangeSDFToClipboard, m.exchanges, enable=bool(m.exchanges)),
            lambda m: m.add(actions.ActivityOpen, [x.input for x in m.exchanges],
                            enable=bool(m.exchanges),
                            text="Open processs" if len(m.exchanges) == 1 else "Open processes",
                            ),
        ]

        @property
        def locked(self):
            return database_is_locked(self.activity["database"])

        @property
        def activity(self):
            return self.parent().activity

        @property
        def exchanges(self):
            indexes = self.parent().selectedIndexes()
            return list(set(idx.internalPointer().exchange for idx in indexes if idx.isValid()))

    def __init__(self, parent):
        """
        Initializes the ExchangesView.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
        """
        super().__init__(parent)
        self.setSortingEnabled(True)

        self.drag_drop_hint = QtWidgets.QLabel("Drag products here to create new exchanges.", self)
        fnt = self.drag_drop_hint.font()
        fnt.setPointSize(fnt.pointSize() + 2)
        fnt.setWeight(QtGui.QFont.Weight.ExtraLight)
        self.drag_drop_hint.setFont(fnt)

        # Set up the layout
        layout = QtWidgets.QVBoxLayout(self)
        layout.addStretch()
        layout.addWidget(self.drag_drop_hint, alignment=Qt.AlignCenter)  # Center horizontally
        layout.addStretch()

        # Set the property delegate
        self.propertyDelegate = delegates.PropertyDelegate(self)

    @property
    def activity(self):
        """
        Returns the activity associated with the view.

        Returns:
            The activity associated with the view.
        """
        return self.parent().activity

    def setDefaultColumnDelegates(self):
        """
        Sets the default column delegates for the view.
        """
        super().setDefaultColumnDelegates()

        columns = self.model().columns()
        for i, col_name in enumerate(columns):
            if not col_name.startswith("property_"):
                continue
            # Set the delegate for property columns
            self.setItemDelegateForColumn(i, self.propertyDelegate)


class ExchangesItem(widgets.ABDataItem):
    """
    An item representing an exchange in the tree view.

    Attributes:
        background_color (str): The background color of the item.
    """
    background_color = None

    @property
    def exchange(self):
        """
        Returns the exchange associated with the item.

        Returns:
            The exchange associated with the item.
        """
        return self["_exchange"]

    @property
    def functional(self):
        """
        Returns whether the exchange is functional.

        Returns:
            bool: True if the exchange is functional, False otherwise.
        """
        return self["_exchange"].get("type") == "production"

    @property
    def scoped_parameters(self):
        """
        Returns the parameters in scope of the current exchange.

        Returns:
            dict: The parameters in scope.
        """
        return bwutils.parameters_in_scope(self["_exchange"].output)

    def flags(self, col: int, key: str):
        """
        Returns the item flags for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the flags.

        Returns:
            QtCore.Qt.ItemFlags: The item flags.
        """
        flags = super().flags(col, key)
        # Check if the database is read-only. If it is, return the default flags.
        if database_is_locked(self.exchange.output["database"]):
            return flags

        # Allow editing for specific keys: "amount", "formula", and "uncertainty".
        if key in ["amount", "formula", "uncertainty", "comment"]:
            return flags | Qt.ItemFlag.ItemIsEditable

        # Allow editing for "unit", "name", "location", and "substitution_factor" if the exchange is functional.
        if key in ["unit", "name", "location", "substitution_factor"] and self.functional:
            return flags | Qt.ItemFlag.ItemIsEditable

        # Allow editing for properties (keys starting with "property_") if the exchange is functional.
        if key.startswith("property_") and self.functional:
            return flags | Qt.ItemFlag.ItemIsEditable

        # Allow editing for "allocation_factor" if the allocation is manual and the exchange is functional.
        if key == "allocation_factor" and self.exchange.output.get("allocation") == "manual" and self.functional:
            return flags | Qt.ItemFlag.ItemIsEditable

        # Return the default flags if none of the above conditions are met.
        return flags

    def displayData(self, col: int, key: str):
        """
        Returns the display data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the display data.

        Returns:
            str: The display data.
        """
        if key in ["allocation_factor", "substitute", "substitution_factor"] and not self.functional:
            return None

        if key.startswith("property_") and not self.functional:
            return None

        if key.startswith("property_") and isinstance(self[key], float):
            return {
                "amount": self[key],
                "unit": "undefined",
                "normalize": False,
            }

        if key.startswith("property_") and self[key].get("normalize", True):
            prop = self[key].copy()
            prop["unit"] = prop['unit'] + f" / {self['unit']}"
            return prop

        return super().displayData(col, key)

    def decorationData(self, col, key):
        """
        Provides decoration data for the item.

        Args:
            col: The column index.
            key: The key for which to provide decoration data.

        Returns:
            The decoration data for the item.
        """
        if key not in ["name", "substitute_name", "amount"] or not self.displayData(col, key):
            return

        if key == "amount":
            if pd.isna(self["formula"]) or self["formula"] is None:
                # empty icon to align the values
                return icons.qicons.empty
            return icons.qicons.parameterized

        if key == "name":
            activity_type = self.exchange.input.get("type")
        else:  # key is "substitute_name"
            activity_type = bd.get_node(key=self["_substitute_key"])["type"]

        if activity_type in ["natural resource", "emission", "inventory indicator", "economic", "social"]:
            return icons.qicons.biosphere
        if activity_type in ["product", "processwithreferenceproduct"]:
            return icons.qicons.product
        if activity_type == "waste":
            return icons.qicons.waste

    def fontData(self, col: int, key: str):
        """
        Returns the font data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the font data.

        Returns:
            QtGui.QFont: The font data.
        """
        font = super().fontData(col, key)

        # set the font to bold if it's a production/functional exchange
        if self.functional:
            font.setWeight(QtGui.QFont.Weight.DemiBold)
        return font

    def backgroundData(self, col: int, key: str):
        """
        Returns the background data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the background data.

        Returns:
            QtGui.QBrush: The background brush for the item.
        """
        if self.background_color:
            return QtGui.QBrush(QtGui.QColor(self.background_color))

        if key == f"property_{self['_allocate_by']}":
            from activity_browser import application
            return application.palette().alternateBase()

    def setData(self, col: int, key: str, value) -> bool:
        """
        Sets the data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to set the data.
            value: The value to set.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if key in ["amount", "formula", "comment"]:
            if key == "formula" and not str(value).strip():
                actions.ExchangeFormulaRemove.run([self.exchange])
                return True

            actions.ExchangeModify.run(self.exchange, {key.lower(): value})
            return True

        if key in ["unit", "name", "location", "substitution_factor", "allocation_factor"]:
            act = self.exchange.input

            # if we're dealing with a legacy activity, we need to set to the product field here
            if key == "name" and not isinstance(act, bf.Product):
                key = "reference product"

            actions.ActivityModify.run(act.key, key.lower(), value)

        if key.startswith("property_"):
            # should move this process to a separate action
            process = self.exchange.output
            product = self.exchange.input

            if not isinstance(process, bf.Process) or not isinstance(product, bf.Product):
                log.warning(f"Expected a Process and Product, got {type(process)} and {type(product)} instead.")
                return False

            prop_key = key[9:]

            prop = process.property_template(prop_key, value)

            props = product.get("properties", {})
            props[prop_key] = prop

            actions.ActivityModify.run(product, "properties", props)

        return False


class ExchangesModel(widgets.ABItemModel):
    """
    A model representing the data for the exchanges.

    Attributes:
        dataItemClass (type): The class of the data items.
    """
    dataItemClass = ExchangesItem




import datetime
import arrow

from qtpy import QtCore, QtGui, QtWidgets


class DateTimeDelegate(QtWidgets.QStyledItemDelegate):
    """For managing and validating entered float values."""

    def displayText(self, value, locale):
        tz = datetime.datetime.now(datetime.timezone.utc).astimezone()
        time_shift = -tz.utcoffset().total_seconds()
        value = arrow.get(value).shift(seconds=time_shift).humanize()
        return value



from typing import TYPE_CHECKING
from qtpy import QtWidgets

if TYPE_CHECKING:
    from activity_browser.ui.widgets import ABWizard
    from activity_browser.ui.threading import ABThread


class ABWizardPage(QtWidgets.QWizardPage):
    title: str = ""
    subtitle: str = ""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle(self.title)
        self.setSubTitle(self.subtitle)

    def wizard(self) -> "ABWizard":
        return super().wizard()

    def nextPage(self) -> type[QtWidgets.QWizardPage] | None:
        i = self.wizard().currentId() + 1
        if i > len(self.wizard().pages) - 1:
            return None
        return self.wizard().pages[i]

    def nextId(self):
        next_page = self.nextPage()
        if next_page is None:
            return -1
        return self.wizard().pages.index(next_page)

    def initializePage(self, context: dict):
        pass

    def finalize(self, context: dict):
        pass


class ABThreadedWizardPage(ABWizardPage):
    Thread: type["ABThread"]

    def __init__(self, parent=None):
        from activity_browser import application

        super().__init__(parent)

        self.thread = self.Thread(application)
        self.thread.finished.connect(self.completeChanged)
        self.thread.finished.connect(lambda: self.statusUpdate(100, "Complete"))
        self.thread.status.connect(self.statusUpdate)

        self.progress_bar = QtWidgets.QProgressBar(self)
        self.progress_bar.setRange(0, 0)
        self.message = QtWidgets.QLabel(self)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.progress_bar)
        layout.addWidget(self.message)

        self.setLayout(layout)

    def statusUpdate(self, progress: int, message: str):
        self.message.setText(message)

        if progress == -1:
            self.progress_bar.setRange(0, 0)
        else:
            self.progress_bar.setRange(0, 100)
            self.progress_bar.setValue(progress)


    def isComplete(self):
        """Check if the download thread has finished"""
        return self.thread.isFinished()




# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets

from activity_browser.ui.icons import qicons


class DeleteButtonDelegate(QtWidgets.QStyledItemDelegate):
    """For showing a persistent delete button in tables"""

    delete_request = QtCore.Signal(QtCore.QModelIndex)

    def __init__(self, parent=None):
        super().__init__(parent)

    def createEditor(self, 
                     parent: QtWidgets.QWidget, 
                     option: QtWidgets.QStyleOptionViewItem, 
                     index: QtCore.QModelIndex):
        button = QtWidgets.QPushButton()
        button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        button.setFixedSize(22, 22)
        button.setAutoDefault(False)
        button.clicked.connect(lambda: self.delete_request.emit(index))
        button.setIcon(qicons.delete)
        editor = QtWidgets.QWidget(parent)
        editor.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(button)
        editor.setLayout(layout)
        return editor



from qtpy import QtWidgets, QtCore

import bw2data as bd


class DatabaseNameEdit(QtWidgets.QWidget):
    """
    Extended LineEdit widget that will check whether the database name provided by the user already exist and warn
    the user accordingly.
    """
    textChanged: QtCore.SignalInstance = QtCore.Signal(str)

    def __init__(self,
                 label: str | None = "",
                 database_placeholder="Database name",
                 database_preset="",
                 overwrite_warning="Existing database will be overwritten",
                 ):
        """
        Parameters
        ----------
            label : `str`
                Header to show above the text field. If an empty string is provided (default), label will not be added
            database_placeholder : `str`
                Text to show in the background of the database field
            database_preset : `str`
                Text with which to fill in the database field as suggestion
            overwrite_warning : `str`
                Text to show as warning when the database already exists. If the string is empty, no warning will be
                shown
        """
        super().__init__()

        self.label = QtWidgets.QLabel(label)

        # Create db name textbox
        self.database_name = QtWidgets.QLineEdit()
        self.database_name.setPlaceholderText(database_placeholder)
        self.database_name.setText(database_preset)
        self.database_name.textChanged.connect(self.textChanged.emit)

        # Only show warning when a string is given
        if overwrite_warning:
            self.database_name.textChanged.connect(self._update)

        # Create warning text for when the user enters a database that already exists
        self.warning = QtWidgets.QLabel()
        self.warning.setTextFormat(QtCore.Qt.RichText)
        self.warning.setText(
            f"<p style='color: red; font-size: small;'>{overwrite_warning}</p>")
        self.warning.setHidden(True)

        layout = QtWidgets.QVBoxLayout()

        if label:
            layout.addWidget(self.label)
        layout.addWidget(self.database_name)
        layout.addWidget(self.warning)

        self.setLayout(layout)

    def _update(self):
        """Slot to check whether the database already exists and show the warning if so"""
        if self.willOverwrite():
            self.warning.setHidden(False)
        else:
            self.warning.setHidden(True)

    def text(self) -> str:
        return self.database_name.text()

    def setText(self, text: str):
        self.database_name.setText(text)

    def willOverwrite(self) -> bool:
        return self.database_name.text() in bd.databases




from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd


class DatabaseSetReadonly(ABAction):
    """
    ABAction to set a database as read-only.

    This action allows marking a database as read-only by updating its metadata.
    It can also be used to remove the read-only status by setting the `read_only` flag to `False`.

    Attributes:
        text (str): The display text for this action.
        tool_tip (str): The tooltip text for this action.
    """

    text = "Set database read-only"
    tool_tip = "Set this database to read-only"

    @staticmethod
    @exception_dialogs
    def run(db_name: str, read_only=True):
        """
        Execute the action to set the read-only status of a database.

        This method updates the `read_only` flag in the metadata of the specified database.

        Args:
            db_name (str): The name of the database to update.
            read_only (bool, optional): The desired read-only status. Defaults to True.
        """
        bd.databases[db_name]["read_only"] = read_only
        bd.databases.flush()
